"""AI summarization for wiki templates."""

from __future__ import annotations

import hashlib
import json
import os
from pathlib import Path
from typing import Any, Dict

from kanbus.models import AiConfiguration

AI_SUMMARIES_CACHE = "ai_summaries.json"
AI_CALLS_LOG = "ai_calls.log"


def _cache_key(issue: Dict[str, Any], detail: str) -> str:
    identifier = str(issue.get("id") or issue.get("identifier", ""))
    updated = str(issue.get("updated_at", ""))
    return hashlib.sha256(f"{identifier}:{updated}:{detail}".encode()).hexdigest()


def _log_call(cache_dir: Path) -> None:
    log_path = cache_dir / AI_CALLS_LOG
    log_path.parent.mkdir(parents=True, exist_ok=True)
    with open(log_path, "a", encoding="utf-8") as f:
        f.write("1\n")


def _read_cache(cache_dir: Path, key: str) -> str | None:
    path = cache_dir / AI_SUMMARIES_CACHE
    if not path.exists():
        return None
    try:
        data = json.loads(path.read_text(encoding="utf-8"))
        return data.get(key)
    except (json.JSONDecodeError, OSError):
        return None


def _write_cache(cache_dir: Path, key: str, value: str) -> None:
    path = cache_dir / AI_SUMMARIES_CACHE
    path.parent.mkdir(parents=True, exist_ok=True)
    data = {}
    if path.exists():
        try:
            data = json.loads(path.read_text(encoding="utf-8"))
        except (json.JSONDecodeError, OSError):
            pass
    data[key] = value
    path.write_text(json.dumps(data, indent=2), encoding="utf-8")


def summarize_issue(
    issue: Dict[str, Any], detail: str, ai_config: AiConfiguration | None
) -> str:
    """Summarize an issue for wiki templates.

    :param issue: Serialized issue dict (from issue() in templates).
    :type issue: Dict[str, Any]
    :param detail: Detail level (e.g. short, medium).
    :type detail: str
    :param ai_config: AI configuration from project, or None.
    :type ai_config: AiConfiguration | None
    :return: Summary text.
    :rtype: str
    """
    if ai_config is None:
        return "(AI summarization not configured)"

    if os.environ.get("KANBUS_TEST_AI_MOCK") == "1":
        identifier = issue.get("id") or issue.get("identifier", "unknown")
        return f"Generated summary for {identifier}"

    return f"Summary: {issue.get('title', 'untitled')}"


def make_ai_summarize(
    issues_by_id: Dict[str, Dict[str, Any]],
    ai_config: AiConfiguration | None,
    cache_dir: Path | None,
):
    """Build ai_summarize callable for wiki template context.

    :param issues_by_id: Map of issue id to serialized issue.
    :type issues_by_id: Dict[str, Dict[str, Any]]
    :param ai_config: AI configuration or None.
    :type ai_config: AiConfiguration | None
    :param cache_dir: Directory for AI cache (project/.cache) or None to skip cache.
    :type cache_dir: Path | None
    :return: Callable(issue_or_id, detail) for templates.
    :rtype: Callable
    """

    def ai_summarize(issue_or_id: Any, detail: str = "short") -> str:
        if isinstance(issue_or_id, dict):
            issue = issue_or_id
        else:
            identifier = str(issue_or_id)
            issue = issues_by_id.get(identifier)
            if issue is None:
                return "(issue not found)"
        if ai_config is None:
            return "(AI summarization not configured)"
        if cache_dir is not None:
            key = _cache_key(issue, detail)
            cached = _read_cache(cache_dir, key)
            if cached is not None:
                return cached
        result = summarize_issue(issue, detail, ai_config)
        if cache_dir is not None:
            key = _cache_key(issue, detail)
            _write_cache(cache_dir, key, result)
            if os.environ.get("KANBUS_TEST_AI_MOCK") == "1":
                _log_call(cache_dir)
        return result

    return ai_summarize



from __future__ import annotations

import json
from datetime import timezone
from pathlib import Path

import pytest

from kanbus import migration
from kanbus.migration import MigrationError

from test_helpers import build_project_configuration


def _record(issue_id: str, **overrides):
    base = {
        "id": issue_id,
        "title": f"Title {issue_id}",
        "description": "desc",
        "issue_type": "task",
        "status": "open",
        "priority": 2,
        "created_at": "2026-03-08T00:00:00Z",
        "updated_at": "2026-03-09T00:00:00Z",
        "dependencies": [],
        "comments": [],
    }
    base.update(overrides)
    return base


def test_load_beads_issues_and_issue_error_paths(tmp_path: Path) -> None:
    with pytest.raises(MigrationError, match="no .beads directory"):
        migration.load_beads_issues(tmp_path)

    beads = tmp_path / ".beads"
    beads.mkdir()
    with pytest.raises(MigrationError, match="no issues.jsonl"):
        migration.load_beads_issues(tmp_path)

    (beads / "issues.jsonl").write_text(
        json.dumps(_record("kanbus-1")) + "\n", encoding="utf-8"
    )
    issues = migration.load_beads_issues(tmp_path)
    assert len(issues) == 1
    assert issues[0].identifier == "kanbus-1"
    assert migration.load_beads_issue(tmp_path, "kanbus-1").identifier == "kanbus-1"

    with pytest.raises(MigrationError, match="not found"):
        migration.load_beads_issue(tmp_path, "missing")


def test_migrate_from_beads_error_paths(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(
        migration,
        "ensure_git_repository",
        lambda _root: (_ for _ in ()).throw(RuntimeError("not a git repo")),
    )
    with pytest.raises(MigrationError, match="not a git repo"):
        migration.migrate_from_beads(tmp_path)

    monkeypatch.setattr(migration, "ensure_git_repository", lambda _root: None)
    with pytest.raises(MigrationError, match="no .beads directory"):
        migration.migrate_from_beads(tmp_path)

    beads = tmp_path / ".beads"
    beads.mkdir()
    with pytest.raises(MigrationError, match="no issues.jsonl"):
        migration.migrate_from_beads(tmp_path)

    (beads / "issues.jsonl").write_text(
        json.dumps(_record("kanbus-1")) + "\n", encoding="utf-8"
    )
    monkeypatch.setattr(
        migration, "discover_project_directories", lambda _root: [Path("project")]
    )
    with pytest.raises(MigrationError, match="already initialized"):
        migration.migrate_from_beads(tmp_path)


def test_migrate_from_beads_success_writes_issues(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    beads = tmp_path / ".beads"
    beads.mkdir()
    records = [_record("kanbus-1"), _record("kanbus-2")]
    (beads / "issues.jsonl").write_text(
        "\n".join(json.dumps(r) for r in records) + "\n", encoding="utf-8"
    )

    project_dir = tmp_path / "project"
    (project_dir / "issues").mkdir(parents=True)

    monkeypatch.setattr(migration, "ensure_git_repository", lambda _root: None)
    monkeypatch.setattr(migration, "discover_project_directories", lambda _root: [])
    monkeypatch.setattr(migration, "initialize_project", lambda _root: None)
    monkeypatch.setattr(
        migration, "get_configuration_path", lambda _root: tmp_path / ".kanbus.yml"
    )
    monkeypatch.setattr(
        migration,
        "load_project_configuration",
        lambda _path: build_project_configuration(),
    )

    writes: list[Path] = []
    monkeypatch.setattr(
        migration, "write_issue_to_file", lambda issue, path: writes.append(path)
    )

    result = migration.migrate_from_beads(tmp_path)
    assert result.issue_count == 2
    assert len(writes) == 2
    assert writes[0].name == "kanbus-1.json"


def test_load_beads_records_and_dedupe(tmp_path: Path) -> None:
    issues_path = tmp_path / "issues.jsonl"
    issues_path.write_text(
        "\n" + json.dumps(_record("kanbus-1")) + "\n", encoding="utf-8"
    )

    records = migration._load_beads_records(issues_path)
    assert len(records) == 1

    issues_path.write_text(json.dumps({"title": "no id"}) + "\n", encoding="utf-8")
    with pytest.raises(MigrationError, match="missing id"):
        migration._load_beads_records(issues_path)

    records = [
        _record("kanbus-1", updated_at="2026-03-09T00:00:00Z"),
        _record("kanbus-1", updated_at="2026-03-09T00:00:00Z", notes="longer payload"),
        _record("kanbus-2"),
        _record("kanbus-2"),
    ]
    deduped = migration._dedupe_beads_records(records, issues_path)
    assert len(deduped) == 2
    assert {r["id"] for r in deduped} == {"kanbus-1", "kanbus-2"}

    with pytest.raises(MigrationError, match="missing id"):
        migration._dedupe_beads_records([{"id": "x"}, {"title": "no id"}], issues_path)

    tied_records = [
        _record("kanbus-3", updated_at="2026-03-09T00:00:00Z", notes="short"),
        _record("kanbus-3", updated_at="2026-03-10T00:00:00Z", notes="long payload"),
        _record(
            "kanbus-3", updated_at="2026-03-10T00:00:00Z", notes="long payload plus"
        ),
    ]
    deduped_tied = migration._dedupe_beads_records(tied_records, issues_path)
    assert len(deduped_tied) == 1
    assert deduped_tied[0]["notes"] == "long payload plus"

    unique_latest_records = [
        _record("kanbus-4", updated_at="2026-03-09T00:00:00Z"),
        _record("kanbus-4", updated_at="2026-03-10T00:00:00Z"),
    ]
    deduped_unique_latest = migration._dedupe_beads_records(
        unique_latest_records, issues_path
    )
    assert len(deduped_unique_latest) == 1
    assert deduped_unique_latest[0]["updated_at"] == "2026-03-10T00:00:00Z"

    class _FlakyIdRecord(dict):
        def __init__(self):
            super().__init__(_record("kanbus-5"))
            self._id_calls = 0

        def get(self, key, default=None):
            if key == "id":
                self._id_calls += 1
                if self._id_calls >= 2:
                    return None
            return super().get(key, default)

    with pytest.raises(MigrationError, match="missing id"):
        migration._dedupe_beads_records([_FlakyIdRecord()], issues_path)


def test_load_configuration_for_beads_builds_expected_defaults(tmp_path: Path) -> None:
    records = [
        _record("kanbus-1", issue_type="feature", status="blocked", priority=4),
        _record("kanbus-2", issue_type="epic", status="open", priority=1),
    ]
    cfg = migration._load_configuration_for_beads(tmp_path, records)
    assert cfg.project_key == "BD"
    assert "story" in cfg.types
    assert "blocked" in {status.key for status in cfg.statuses}
    assert 4 in cfg.priorities


def test_convert_record_and_validations(monkeypatch: pytest.MonkeyPatch) -> None:
    cfg = build_project_configuration()
    record_by_id = {"kanbus-parent": _record("kanbus-parent", issue_type="epic")}

    full = _record(
        "kanbus-1",
        issue_type="feature",
        dependencies=[{"type": "parent-child", "depends_on_id": "kanbus-parent"}],
        comments=[
            {"id": 1, "author": "A", "text": "Hi", "created_at": "2026-03-09T00:00:00Z"}
        ],
        owner="owner",
        notes="notes",
        acceptance_criteria="ac",
        close_reason="done",
        created_by="creator",
        assignee="assignee",
        labels=["x"],
    )
    issue = migration._convert_record(full, record_by_id, cfg)
    assert issue.identifier == "kanbus-1"
    assert issue.issue_type == "story"
    assert issue.parent == "kanbus-parent"
    assert issue.custom["beads_owner"] == "owner"
    assert issue.custom["beads_issue_type"] == "feature"

    with_closed_at = _record("kanbus-closed", closed_at="2026-03-09T01:00:00Z")
    issue_closed = migration._convert_record(with_closed_at, record_by_id, cfg)
    assert issue_closed.closed_at is not None

    for key, value in [
        ("title", ""),
        ("issue_type", ""),
        ("status", ""),
    ]:
        bad = _record("bad", **{key: value})
        with pytest.raises(MigrationError):
            migration._convert_record(bad, record_by_id, cfg)

    with pytest.raises(MigrationError, match="priority is required"):
        migration._convert_record(_record("badp", priority=None), record_by_id, cfg)

    with pytest.raises(MigrationError, match="invalid priority"):
        migration._convert_record(_record("badp2", priority=99), record_by_id, cfg)


def test_convert_dependencies_paths(monkeypatch: pytest.MonkeyPatch) -> None:
    cfg = build_project_configuration()
    record_by_id = {
        "p1": _record("p1", issue_type="epic"),
        "p2": _record("p2", issue_type="epic"),
        "d1": _record("d1", issue_type="task"),
    }

    warnings: list[str] = []
    monkeypatch.setattr(
        migration.click, "echo", lambda msg, err=True: warnings.append(msg)
    )

    parent, deps = migration._convert_dependencies(
        [
            {"type": "parent-child", "depends_on_id": "p1"},
            {"type": "parent-child", "depends_on_id": "p2"},
            {"type": "blocks", "depends_on_id": "d1"},
        ],
        "kanbus-1",
        record_by_id,
        cfg,
        "task",
    )
    assert parent == "p1"
    assert len(deps) == 1
    assert deps[0].target == "d1"
    assert any("multiple parents" in w for w in warnings)

    with pytest.raises(MigrationError, match="invalid dependency"):
        migration._convert_dependencies(
            [{"type": "blocks"}], "x", record_by_id, cfg, "task"
        )

    with pytest.raises(MigrationError, match="missing dependency"):
        migration._convert_dependencies(
            [{"type": "blocks", "depends_on_id": "nope"}],
            "x",
            record_by_id,
            cfg,
            "task",
        )

    with pytest.raises(MigrationError, match="parent issue_type is required"):
        migration._convert_dependencies(
            [{"type": "parent-child", "depends_on_id": "missing-type"}],
            "x",
            {"missing-type": {"id": "missing-type"}},
            cfg,
            "task",
        )


def test_convert_dependencies_invalid_hierarchy_suggests_and_drops_parent(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cfg = build_project_configuration()
    record_by_id = {"p": _record("p", issue_type="story")}
    warnings: list[str] = []
    monkeypatch.setattr(
        migration.click, "echo", lambda msg, err=True: warnings.append(msg)
    )

    parent, deps = migration._convert_dependencies(
        [{"type": "parent-child", "depends_on_id": "p"}],
        "kanbus-1",
        record_by_id,
        cfg,
        "epic",
    )
    assert parent is None
    assert deps == []
    assert any("Suggestion:" in w for w in warnings)


def test_comment_conversion_and_uuid() -> None:
    uuid_a = migration._beads_comment_uuid("kanbus-1", "1")
    uuid_b = migration._beads_comment_uuid("kanbus-1", "1")
    assert uuid_a == uuid_b

    comment = migration._convert_comment(
        "kanbus-1",
        {"id": 7, "author": "A", "text": "B", "created_at": "2026-03-09T00:00:00Z"},
        0,
    )
    assert comment.author == "A"
    assert comment.text == "B"

    with pytest.raises(MigrationError, match="invalid comment"):
        migration._convert_comment(
            "kanbus-1",
            {"author": "", "text": "x", "created_at": "2026-03-09T00:00:00Z"},
            0,
        )


def test_timestamp_normalization_and_validation_helpers() -> None:
    parsed = migration._parse_timestamp("2026-03-09T00:00:00Z", "updated_at")
    assert parsed.tzinfo == timezone.utc

    parsed_naive = migration._parse_timestamp("2026-03-09T00:00:00", "updated_at")
    assert parsed_naive.tzinfo == timezone.utc

    assert (
        migration._normalize_fractional_seconds("2026-03-09T00:00:00Z")
        == "2026-03-09T00:00:00Z"
    )
    assert (
        migration._normalize_fractional_seconds("2026-03-09T00:00:00.1+00:00")
        == "2026-03-09T00:00:00.100000+00:00"
    )
    assert (
        migration._normalize_fractional_seconds("2026-03-09T00:00:00.123456789+00:00")
        == "2026-03-09T00:00:00.123456+00:00"
    )
    assert (
        migration._normalize_fractional_seconds("2026-03-09T00:00:00.ab+00:00")
        == "2026-03-09T00:00:00.ab+00:00"
    )
    assert (
        migration._normalize_fractional_seconds("2026-03-09T00:00:00.123456")
        == "2026-03-09T00:00:00.123456"
    )

    with pytest.raises(MigrationError, match="updated_at is required"):
        migration._parse_timestamp(None, "updated_at")
    with pytest.raises(MigrationError, match="updated_at must be a string"):
        migration._parse_timestamp(123, "updated_at")
    with pytest.raises(MigrationError, match="invalid updated_at"):
        migration._parse_timestamp("bad", "updated_at")

    cfg = build_project_configuration()
    migration._validate_issue_type(cfg, "task")
    with pytest.raises(MigrationError, match="unknown issue type"):
        migration._validate_issue_type(cfg, "unknown")

    migration._validate_status(cfg, "task", "open")
    with pytest.raises(MigrationError, match="invalid status"):
        migration._validate_status(cfg, "task", "nope")



"""
Project configuration defaults and writers.
"""

from __future__ import annotations

from typing import Any, Dict, List

DEFAULT_HIERARCHY: List[str] = ["initiative", "epic", "task", "sub-task"]
DEFAULT_TYPES: List[str] = ["bug", "story", "chore"]

DEFAULT_CONFIGURATION: Dict[str, Any] = {
    "project_directory": "project",
    "virtual_projects": {},
    "console_port": None,
    "realtime": {
        "transport": "auto",
        "broker": "auto",
        "autostart": True,
        "keepalive": False,
        "uds_socket_path": None,
        "topics": {
            "project_events": "projects/{project}/events",
        },
    },
    "overlay": {
        "enabled": True,
        "ttl_s": 86400,
    },
    "hooks": {
        "enabled": True,
        "run_in_beads_mode": True,
        "default_timeout_ms": 5000,
        "before": {},
        "after": {},
    },
    "project_key": "kanbus",
    "hierarchy": DEFAULT_HIERARCHY,
    "types": DEFAULT_TYPES,
    "workflows": {
        "default": {
            "backlog": ["open", "closed"],
            "open": ["in_progress", "closed", "backlog"],
            "in_progress": ["open", "blocked", "closed"],
            "blocked": ["in_progress", "closed"],
            "closed": ["open"],
        },
        "epic": {
            "open": ["in_progress", "closed"],
            "in_progress": ["open", "closed"],
            "closed": ["open"],
        },
    },
    "initial_status": "open",
    "priorities": {
        0: {"name": "critical", "color": "red"},
        1: {"name": "high", "color": "bright_red"},
        2: {"name": "medium", "color": "yellow"},
        3: {"name": "low", "color": "blue"},
        4: {"name": "trivial", "color": "white"},
    },
    "default_priority": 2,
    "assignee": None,
    "time_zone": None,
    "categories": [
        {"name": "To do", "color": "grey"},
        {"name": "In progress", "color": "blue"},
        {"name": "Done", "color": "green"},
    ],
    "sort_order": {},
    "statuses": [
        {"key": "backlog", "name": "Backlog", "category": "To do", "collapsed": True},
        {"key": "open", "name": "Discovery", "category": "To do", "collapsed": False},
        {
            "key": "in_progress",
            "name": "In Progress",
            "category": "In progress",
            "collapsed": False,
        },
        {
            "key": "blocked",
            "name": "Blocked",
            "category": "In progress",
            "collapsed": True,
        },
        {"key": "closed", "name": "Done", "category": "Done", "collapsed": True},
    ],
    "transition_labels": {
        "default": {
            "backlog": {
                "open": "Start discovery",
                "closed": "Drop",
            },
            "open": {
                "in_progress": "Start work",
                "closed": "Drop",
                "backlog": "Back to backlog",
            },
            "in_progress": {
                "open": "Pause",
                "blocked": "Block",
                "closed": "Complete",
            },
            "blocked": {"in_progress": "Unblock", "closed": "Drop"},
            "closed": {"open": "Reopen"},
        },
        "epic": {
            "open": {"in_progress": "Start", "closed": "Complete"},
            "in_progress": {"open": "Pause", "closed": "Complete"},
            "closed": {"open": "Reopen"},
        },
    },
    "type_colors": {
        "initiative": "bright_blue",
        "epic": "magenta",
        "task": "blue",
        "sub-task": "bright_cyan",
        "bug": "red",
        "story": "yellow",
        "chore": "green",
        "event": "bright_blue",
    },
    "beads_compatibility": False,
    "wiki_directory": None,
    "github_security": None,
}



"""In-memory index building for Kanbus issues."""

from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass, field
import json
import os
from pathlib import Path
from typing import Dict, List

from kanbus.models import IssueData


def _load_issue_data(issue_path: Path) -> IssueData:
    """Load a single issue JSON file into an IssueData model.

    :param issue_path: Path to the issue JSON file.
    :type issue_path: Path
    :return: Parsed issue model.
    :rtype: IssueData
    """
    payload = json.loads(issue_path.read_bytes())
    return IssueData.model_validate(payload)


def _load_issue_batch(issue_paths: List[Path]) -> List[IssueData]:
    """Load a batch of issue JSON files into IssueData models.

    :param issue_paths: List of issue JSON paths.
    :type issue_paths: List[Path]
    :return: Parsed issue models.
    :rtype: List[IssueData]
    """
    return [_load_issue_data(issue_path) for issue_path in issue_paths]


def _add_issue_to_index(index: IssueIndex, issue: IssueData) -> None:
    """Add an IssueData instance to all index lookup tables.

    :param index: Index to update.
    :type index: IssueIndex
    :param issue: Issue data to register.
    :type issue: IssueData
    """
    index.by_id[issue.identifier] = issue
    index.by_status.setdefault(issue.status, []).append(issue)
    index.by_type.setdefault(issue.issue_type, []).append(issue)
    if issue.parent is not None:
        index.by_parent.setdefault(issue.parent, []).append(issue)
    for label in issue.labels:
        index.by_label.setdefault(label, []).append(issue)
    for dependency in issue.dependencies:
        if dependency.dependency_type == "blocked-by":
            index.reverse_dependencies.setdefault(dependency.target, []).append(issue)


@dataclass
class IssueIndex:
    """In-memory lookup tables for issues."""

    by_id: Dict[str, IssueData] = field(default_factory=dict)
    by_status: Dict[str, List[IssueData]] = field(default_factory=dict)
    by_type: Dict[str, List[IssueData]] = field(default_factory=dict)
    by_parent: Dict[str, List[IssueData]] = field(default_factory=dict)
    by_label: Dict[str, List[IssueData]] = field(default_factory=dict)
    reverse_dependencies: Dict[str, List[IssueData]] = field(default_factory=dict)


def build_index_from_directory(issues_directory: Path) -> IssueIndex:
    """Build an IssueIndex by scanning issue files in a directory.

    :param issues_directory: Directory containing issue JSON files.
    :type issues_directory: Path
    :return: In-memory issue index.
    :rtype: IssueIndex
    """
    index = IssueIndex()
    issue_entries = [
        (entry.name, entry.path)
        for entry in os.scandir(issues_directory)
        if entry.is_file() and entry.name.endswith(".json")
    ]
    issue_entries.sort(key=lambda item: item[0])
    issue_paths = [Path(issue_path) for _name, issue_path in issue_entries]
    if issue_paths:
        max_workers = min(4, len(issue_paths))
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            chunk_size = (len(issue_paths) + max_workers - 1) // max_workers
            chunks = [
                issue_paths[index : index + chunk_size]
                for index in range(0, len(issue_paths), chunk_size)
            ]
            batches = executor.map(_load_issue_batch, chunks)
            issues = [issue for batch in batches for issue in batch]
        for issue in issues:
            _add_issue_to_index(index, issue)
    return index



from __future__ import annotations

from pathlib import Path
from types import SimpleNamespace

import pytest

from kanbus import issue_creation
from kanbus.hierarchy import InvalidHierarchyError
from kanbus.issue_lookup import IssueLookupError
from kanbus.workflows import InvalidTransitionError

from test_helpers import build_issue, build_project_configuration


def _setup_common(
    monkeypatch: pytest.MonkeyPatch, root: Path, *, local_dir: Path | None = None
):
    project_dir = root / "project"
    issues_dir = project_dir / "issues"
    issues_dir.mkdir(parents=True, exist_ok=True)
    cfg = build_project_configuration(project_directory="project")

    monkeypatch.setattr(
        issue_creation, "load_project_directory", lambda _r: project_dir
    )
    monkeypatch.setattr(
        issue_creation, "find_project_local_directory", lambda _p: local_dir
    )
    monkeypatch.setattr(
        issue_creation, "get_configuration_path", lambda _r: root / ".kanbus.yml"
    )
    monkeypatch.setattr(issue_creation, "load_project_configuration", lambda _p: cfg)
    monkeypatch.setattr(issue_creation, "list_issue_identifiers", lambda _d: set())
    monkeypatch.setattr(
        issue_creation,
        "generate_issue_identifier",
        lambda _req: SimpleNamespace(identifier="kanbus-new"),
    )
    monkeypatch.setattr(issue_creation, "get_current_user", lambda: "dev")
    monkeypatch.setattr(
        issue_creation, "now_timestamp", lambda: "2026-03-09T00:00:00.000Z"
    )
    monkeypatch.setattr(
        issue_creation,
        "create_event",
        lambda **_kwargs: SimpleNamespace(event_id="evt-1"),
    )
    monkeypatch.setattr(
        issue_creation, "issue_created_payload", lambda _issue: {"ok": True}
    )
    return project_dir, issues_dir, cfg


def test_create_issue_success_non_local(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root = tmp_path
    project_dir, _issues_dir, _cfg = _setup_common(monkeypatch, root)

    written: dict[str, object] = {}
    monkeypatch.setattr(
        issue_creation,
        "write_issue_to_file",
        lambda issue, path: written.update({"issue": issue, "path": path}),
    )
    monkeypatch.setattr(
        issue_creation, "events_dir_for_project", lambda _p: project_dir / "events"
    )
    monkeypatch.setattr(issue_creation, "write_events_batch", lambda *_a: None)
    published: list[str] = []
    monkeypatch.setattr(
        issue_creation, "publish_issue_mutation", lambda *_a: published.append("pub")
    )

    result = issue_creation.create_issue(
        root,
        title="Hello",
        issue_type=None,
        priority=None,
        assignee=None,
        parent=None,
        labels=["a"],
        description="desc",
        local=False,
        validate=False,
    )
    assert result.issue.identifier == "kanbus-new"
    assert written["path"] == project_dir / "issues" / "kanbus-new.json"
    assert published == ["pub"]


def test_create_issue_success_local_uses_local_events(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root = tmp_path
    local_dir = root / "project-local"
    local_issues = local_dir / "issues"
    local_issues.mkdir(parents=True)
    project_dir, _issues_dir, _cfg = _setup_common(
        monkeypatch, root, local_dir=local_dir
    )

    monkeypatch.setattr(
        issue_creation, "ensure_project_local_directory", lambda _p: local_dir
    )
    target_paths: list[Path] = []
    monkeypatch.setattr(
        issue_creation,
        "write_issue_to_file",
        lambda _issue, path: target_paths.append(path),
    )
    monkeypatch.setattr(
        issue_creation,
        "events_dir_for_local",
        lambda _p: project_dir / "project-local-events",
    )
    monkeypatch.setattr(issue_creation, "write_events_batch", lambda *_a: None)
    monkeypatch.setattr(
        issue_creation,
        "publish_issue_mutation",
        lambda *_a: (_ for _ in ()).throw(RuntimeError("should not publish")),
    )

    result = issue_creation.create_issue(
        root,
        title="Hello",
        issue_type=None,
        priority=None,
        assignee=None,
        parent=None,
        labels=[],
        description=None,
        local=True,
        validate=False,
    )
    assert result.issue.identifier == "kanbus-new"
    assert target_paths == [local_issues / "kanbus-new.json"]


def test_create_issue_wraps_project_or_config_errors(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(
        issue_creation,
        "load_project_directory",
        lambda _r: (_ for _ in ()).throw(issue_creation.ProjectMarkerError("bad root")),
    )
    with pytest.raises(issue_creation.IssueCreationError, match="bad root"):
        issue_creation.create_issue(tmp_path, "t", None, None, None, None, [], None)

    project_dir = tmp_path / "project"
    project_dir.mkdir(parents=True)
    monkeypatch.setattr(
        issue_creation, "load_project_directory", lambda _r: project_dir
    )
    monkeypatch.setattr(issue_creation, "find_project_local_directory", lambda _p: None)
    monkeypatch.setattr(
        issue_creation, "get_configuration_path", lambda _r: tmp_path / ".kanbus.yml"
    )
    monkeypatch.setattr(
        issue_creation,
        "load_project_configuration",
        lambda _p: (_ for _ in ()).throw(issue_creation.ConfigurationError("bad cfg")),
    )
    with pytest.raises(issue_creation.IssueCreationError, match="bad cfg"):
        issue_creation.create_issue(tmp_path, "t", None, None, None, None, [], None)


def test_create_issue_validation_errors(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root = tmp_path
    _project_dir, issues_dir, cfg = _setup_common(monkeypatch, root)
    monkeypatch.setattr(issue_creation, "write_issue_to_file", lambda *_a: None)
    monkeypatch.setattr(
        issue_creation, "events_dir_for_project", lambda _p: root / "events"
    )
    monkeypatch.setattr(issue_creation, "write_events_batch", lambda *_a: None)
    monkeypatch.setattr(issue_creation, "publish_issue_mutation", lambda *_a: None)

    with pytest.raises(issue_creation.IssueCreationError, match="unknown issue type"):
        issue_creation.create_issue(
            root, "t", "invalid", None, None, None, [], None, validate=True
        )

    with pytest.raises(issue_creation.IssueCreationError, match="invalid priority"):
        issue_creation.create_issue(
            root, "t", "task", 999, None, None, [], None, validate=True
        )

    monkeypatch.setattr(
        issue_creation,
        "resolve_issue_identifier",
        lambda *_a: (_ for _ in ()).throw(IssueLookupError("parent missing")),
    )
    with pytest.raises(issue_creation.IssueCreationError, match="parent missing"):
        issue_creation.create_issue(
            root, "t", "task", None, None, "p", [], None, validate=True
        )

    monkeypatch.setattr(
        issue_creation, "resolve_issue_identifier", lambda *_a: "kanbus-parent"
    )
    with pytest.raises(issue_creation.IssueCreationError, match="not found"):
        issue_creation.create_issue(
            root, "t", "task", None, None, "p", [], None, validate=True
        )

    parent_path = issues_dir / "kanbus-parent.json"
    parent_path.write_text("{}", encoding="utf-8")
    monkeypatch.setattr(
        issue_creation,
        "read_issue_from_file",
        lambda _p: build_issue("kanbus-parent", issue_type="epic"),
    )
    monkeypatch.setattr(
        issue_creation,
        "validate_parent_child_relationship",
        lambda *_a: (_ for _ in ()).throw(InvalidHierarchyError("bad hierarchy")),
    )
    with pytest.raises(issue_creation.IssueCreationError, match="bad hierarchy"):
        issue_creation.create_issue(
            root, "t", "task", None, None, "p", [], None, validate=True
        )

    monkeypatch.setattr(
        issue_creation, "validate_parent_child_relationship", lambda *_a: None
    )
    monkeypatch.setattr(
        issue_creation, "_find_duplicate_title", lambda *_a: "kanbus-dup"
    )
    with pytest.raises(issue_creation.IssueCreationError, match="duplicate title"):
        issue_creation.create_issue(
            root, "t", "task", None, None, None, [], None, validate=True
        )

    monkeypatch.setattr(issue_creation, "_find_duplicate_title", lambda *_a: None)
    monkeypatch.setattr(
        issue_creation,
        "validate_status_value",
        lambda *_a: (_ for _ in ()).throw(InvalidTransitionError("bad status")),
    )
    with pytest.raises(issue_creation.IssueCreationError, match="bad status"):
        issue_creation.create_issue(
            root, "t", "task", None, None, None, [], None, validate=True
        )


def test_create_issue_policy_violation_and_event_rollback(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root = tmp_path
    project_dir, issues_dir, cfg = _setup_common(monkeypatch, root)
    monkeypatch.setattr(issue_creation, "validate_status_value", lambda *_a: None)
    monkeypatch.setattr(issue_creation, "_find_duplicate_title", lambda *_a: None)
    monkeypatch.setattr(
        issue_creation,
        "write_issue_to_file",
        lambda issue, path: path.write_text(issue.model_dump_json(), encoding="utf-8"),
    )
    monkeypatch.setattr(
        issue_creation, "events_dir_for_project", lambda _p: project_dir / "events"
    )
    monkeypatch.setattr(
        issue_creation,
        "write_events_batch",
        lambda *_a: (_ for _ in ()).throw(RuntimeError("event write fail")),
    )
    monkeypatch.setattr(issue_creation, "publish_issue_mutation", lambda *_a: None)

    policies_dir = project_dir / "policies"
    policies_dir.mkdir(parents=True, exist_ok=True)

    import kanbus.policy_loader as policy_loader
    import kanbus.policy_evaluator as policy_evaluator
    import kanbus.issue_listing as issue_listing

    monkeypatch.setattr(policy_loader, "load_policies", lambda _p: [("x", object())])
    monkeypatch.setattr(issue_listing, "load_issues_from_directory", lambda _d: [])

    import kanbus.policy_context as policy_context

    monkeypatch.setattr(
        policy_evaluator,
        "evaluate_policies",
        lambda *_a: (_ for _ in ()).throw(
            policy_context.PolicyViolationError(
                "policy.policy",
                "Scenario",
                "Given something",
                "policy violated",
                "kanbus-new",
            )
        ),
    )

    with pytest.raises(issue_creation.IssueCreationError, match="policy violated"):
        issue_creation.create_issue(
            root, "t", "task", None, None, None, [], None, validate=True
        )

    monkeypatch.setattr(policy_loader, "load_policies", lambda _p: [])

    with pytest.raises(issue_creation.IssueCreationError, match="event write fail"):
        issue_creation.create_issue(
            root, "t", "task", None, None, None, [], None, validate=True
        )

    assert not (issues_dir / "kanbus-new.json").exists()


def test_find_duplicate_title_handles_invalid_and_casefold(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    issues_dir = tmp_path / "issues"
    issues_dir.mkdir()
    (issues_dir / "a.json").write_text("{}", encoding="utf-8")
    (issues_dir / "b.json").write_text("{}", encoding="utf-8")

    calls = {"n": 0}

    def fake_read(path: Path):
        calls["n"] += 1
        if path.name == "a.json":
            raise ValueError("bad")
        return build_issue("kanbus-2", title="  HeLLo  ")

    monkeypatch.setattr(issue_creation, "read_issue_from_file", fake_read)
    duplicate = issue_creation._find_duplicate_title(issues_dir, "hello")
    assert duplicate == "kanbus-2"

    monkeypatch.setattr(
        issue_creation,
        "read_issue_from_file",
        lambda _p: build_issue("x", title="other"),
    )
    assert issue_creation._find_duplicate_title(issues_dir, "hello") is None



"""Default wiki page content for new projects."""

from __future__ import annotations

DEFAULT_WIKI_INDEX_FILENAME = "index.md"
DEFAULT_WIKI_WHATS_NEXT_FILENAME = "whats-next.md"

DEFAULT_WIKI_INDEX = """# Project Wiki

Welcome to the project wiki.

## What's Next

See [What's Next](whats-next.md) for a report of upcoming work.
"""

DEFAULT_WIKI_WHATS_NEXT = """# What's Next

Open issues, ordered by priority:

{% for issue in query(status="open", sort="priority") %}
- [{{ issue.title }}]({{ issue.id }})
{% endfor %}

{% if count(status="open") == 0 %}
No open issues.
{% endif %}
"""



"""Maintenance command implementations."""

from __future__ import annotations

import json
from pathlib import Path
from dataclasses import dataclass
from typing import Dict, List, Optional, Set

from pydantic import ValidationError as PydanticValidationError

from kanbus.config_loader import ConfigurationError, load_project_configuration
from kanbus.dependencies import ALLOWED_DEPENDENCY_TYPES
from kanbus.hierarchy import InvalidHierarchyError, validate_parent_child_relationship
from kanbus.models import IssueData, ProjectConfiguration
from kanbus.project import (
    ProjectMarkerError,
    get_configuration_path,
    load_project_directory,
)
from kanbus.workflows import get_workflow_for_issue_type


class ProjectValidationError(RuntimeError):
    """Raised when project validation fails."""


class ProjectStatsError(RuntimeError):
    """Raised when project statistics cannot be computed."""


@dataclass(frozen=True)
class ProjectStats:
    """Aggregate issue statistics for a project.

    :param total: Total number of issues.
    :type total: int
    :param open_count: Number of open issues.
    :type open_count: int
    :param closed_count: Number of closed issues.
    :type closed_count: int
    :param type_counts: Counts by issue type.
    :type type_counts: Dict[str, int]
    """

    total: int
    open_count: int
    closed_count: int
    type_counts: Dict[str, int]


def validate_project(root: Path) -> None:
    """Validate issue data and configuration for a Kanbus project.

    :param root: Repository root path.
    :type root: Path
    :raises ProjectValidationError: If validation fails.
    """
    try:
        project_dir = load_project_directory(root)
    except ProjectMarkerError as error:
        raise ProjectValidationError(str(error)) from error

    issues_dir = project_dir / "issues"
    if not issues_dir.exists():
        raise ProjectValidationError("issues directory missing")

    try:
        configuration = load_project_configuration(get_configuration_path(project_dir))
    except ConfigurationError as error:
        raise ProjectValidationError(str(error)) from error

    errors: List[str] = []
    issues: Dict[str, IssueData] = {}
    for issue_path in sorted(issues_dir.glob("*.json"), key=lambda path: path.name):
        issue = _load_issue(issue_path, errors)
        if issue is None:
            continue
        if issue.identifier in issues:
            errors.append(f"{issue_path.name}: duplicate issue id '{issue.identifier}'")
            continue
        _validate_issue_fields(issue_path.name, issue, configuration, errors)
        issues[issue.identifier] = issue

    _validate_references(issues, configuration, errors)

    if errors:
        raise ProjectValidationError(_format_errors(errors))


def collect_project_stats(root: Path) -> ProjectStats:
    """Collect project statistics from issue data.

    :param root: Repository root path.
    :type root: Path
    :return: Aggregated project statistics.
    :rtype: ProjectStats
    :raises ProjectStatsError: If stats cannot be computed.
    """
    try:
        project_dir = load_project_directory(root)
    except ProjectMarkerError as error:
        raise ProjectStatsError(str(error)) from error

    issues_dir = project_dir / "issues"
    if not issues_dir.exists():
        raise ProjectStatsError("issues directory missing")

    issues: List[IssueData] = []
    for issue_path in sorted(issues_dir.glob("*.json"), key=lambda path: path.name):
        try:
            payload = json.loads(issue_path.read_text(encoding="utf-8"))
        except json.JSONDecodeError as error:
            raise ProjectStatsError(
                f"{issue_path.name}: invalid json: {error}"
            ) from error
        try:
            issues.append(IssueData.model_validate(payload))
        except PydanticValidationError as error:
            raise ProjectStatsError(
                f"{issue_path.name}: invalid issue data: {error}"
            ) from error

    total = len(issues)
    closed_count = sum(1 for issue in issues if issue.status == "closed")
    open_count = total - closed_count
    type_counts: Dict[str, int] = {}
    for issue in issues:
        type_counts[issue.issue_type] = type_counts.get(issue.issue_type, 0) + 1

    return ProjectStats(
        total=total,
        open_count=open_count,
        closed_count=closed_count,
        type_counts=type_counts,
    )


def _load_issue(issue_path: Path, errors: List[str]) -> Optional[IssueData]:
    try:
        payload = json.loads(issue_path.read_text(encoding="utf-8"))
    except OSError as error:
        errors.append(f"{issue_path.name}: unable to read issue: {error}")
        return None
    except json.JSONDecodeError as error:
        errors.append(f"{issue_path.name}: invalid json: {error}")
        return None
    try:
        return IssueData.model_validate(payload)
    except PydanticValidationError as error:
        errors.append(f"{issue_path.name}: invalid issue data: {error}")
        return None


def _validate_issue_fields(
    filename: str,
    issue: IssueData,
    configuration: ProjectConfiguration,
    errors: List[str],
) -> None:
    valid_types = configuration.hierarchy + configuration.types
    if issue.identifier != Path(filename).stem:
        errors.append(
            f"{filename}: issue id '{issue.identifier}' does not match filename"
        )
    if issue.issue_type not in valid_types:
        errors.append(f"{filename}: unknown issue type '{issue.issue_type}'")
    if issue.priority not in configuration.priorities:
        errors.append(f"{filename}: invalid priority '{issue.priority}'")

    statuses = _collect_workflow_statuses(configuration, issue.issue_type, errors)
    if statuses is not None and issue.status not in statuses:
        errors.append(f"{filename}: invalid status '{issue.status}'")

    if issue.status == "closed" and issue.closed_at is None:
        errors.append(f"{filename}: closed issues must have closed_at set")
    if issue.status != "closed" and issue.closed_at is not None:
        errors.append(f"{filename}: non-closed issues must not set closed_at")

    for dependency in issue.dependencies:
        if dependency.dependency_type not in ALLOWED_DEPENDENCY_TYPES:
            errors.append(
                f"{filename}: invalid dependency type '{dependency.dependency_type}'"
            )


def _collect_workflow_statuses(
    configuration: ProjectConfiguration,
    issue_type: str,
    errors: List[str],
) -> Optional[Set[str]]:
    try:
        workflow = get_workflow_for_issue_type(configuration, issue_type)
    except ValueError as error:
        errors.append(str(error))
        return None
    statuses: Set[str] = set(workflow.keys())
    for transitions in workflow.values():
        statuses.update(transitions)
    return statuses


def _validate_references(
    issues: Dict[str, IssueData],
    configuration: ProjectConfiguration,
    errors: List[str],
) -> None:
    for issue in issues.values():
        if issue.parent:
            parent = issues.get(issue.parent)
            if parent is None:
                errors.append(
                    f"{issue.identifier}: parent '{issue.parent}' does not exist"
                )
            else:
                try:
                    validate_parent_child_relationship(
                        configuration, parent.issue_type, issue.issue_type
                    )
                except InvalidHierarchyError as error:
                    errors.append(f"{issue.identifier}: {error}")

        for dependency in issue.dependencies:
            if dependency.target not in issues:
                errors.append(
                    f"{issue.identifier}: dependency target '{dependency.target}' does not exist"
                )


def _format_errors(errors: List[str]) -> str:
    return "validation failed:\n" + "\n".join(errors)



"""Issue lookup helpers for project directories."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from kanbus.ids import matches_issue_identifier
from kanbus.config_loader import ConfigurationError, load_project_configuration
from kanbus.issue_files import list_issue_identifiers, read_issue_from_file
from kanbus.models import IssueData, OverlayConfig, ProjectConfiguration
from kanbus.overlay import (
    load_overlay_issue,
    load_tombstone,
    resolve_issue_with_overlay,
)
from kanbus.project import (
    ProjectMarkerError,
    discover_project_directories,
    find_project_local_directory,
    get_configuration_path,
    resolve_labeled_projects,
)


class IssueLookupError(RuntimeError):
    """Raised when an issue lookup fails."""


@dataclass(frozen=True)
class IssueLookupResult:
    """Result of issue lookup."""

    issue: IssueData
    issue_path: Path
    project_dir: Path


def load_issue_from_project(root: Path, identifier: str) -> IssueLookupResult:
    """Load an issue by identifier from a project directory.

    Searches all project directories (including virtual projects and local
    directories) for the given identifier.

    :param root: Repository root path.
    :type root: Path
    :param identifier: Issue identifier.
    :type identifier: str
    :return: Issue lookup result.
    :rtype: IssueLookupResult
    :raises IssueLookupError: If the issue cannot be found.
    """
    try:
        project_dirs = discover_project_directories(root)
    except ProjectMarkerError as error:
        raise IssueLookupError(str(error)) from error

    if not project_dirs:
        raise IssueLookupError("project not initialized")

    all_matches: list[tuple[str, Path, Path]] = []
    root_configuration: ProjectConfiguration | None = None
    project_labels: dict[Path, str] = {}
    try:
        root_configuration = load_project_configuration(get_configuration_path(root))
        project_labels = {
            project.project_dir: project.label
            for project in resolve_labeled_projects(root)
        }
    except (ProjectMarkerError, ConfigurationError):
        root_configuration = None

    overlay_configs = {
        project_dir: _overlay_config_for_project(project_dir, root_configuration)
        for project_dir in project_dirs
    }

    for project_dir in project_dirs:
        overlay_config = overlay_configs.get(project_dir, OverlayConfig(enabled=False))
        for issues_dir in _search_directories(project_dir):
            issue_path = issues_dir / f"{identifier}.json"
            if issue_path.exists():
                issue = read_issue_from_file(issue_path)
                if issues_dir == project_dir / "issues":
                    overlay_issue = load_overlay_issue(project_dir, issue.identifier)
                    tombstone = load_tombstone(project_dir, issue.identifier)
                    issue = resolve_issue_with_overlay(
                        project_dir,
                        issue,
                        overlay_issue,
                        tombstone,
                        overlay_config,
                        project_label=project_labels.get(project_dir),
                    )
                    if issue is None:
                        continue
                return IssueLookupResult(
                    issue=issue, issue_path=issue_path, project_dir=project_dir
                )

            matches = _find_matching_issues(issues_dir, identifier)
            for full_id, path in matches:
                all_matches.append((full_id, path, project_dir))

        overlay_issue = load_overlay_issue(project_dir, identifier)
        if overlay_issue is not None:
            tombstone = load_tombstone(project_dir, identifier)
            resolved = resolve_issue_with_overlay(
                project_dir,
                None,
                overlay_issue,
                tombstone,
                overlay_config,
                project_label=project_labels.get(project_dir),
            )
            if resolved is not None:
                overlay_path = (
                    project_dir / ".overlay" / "issues" / f"{identifier}.json"
                )
                return IssueLookupResult(
                    issue=resolved,
                    issue_path=overlay_path,
                    project_dir=project_dir,
                )

    if not all_matches:
        raise IssueLookupError("not found")
    if len(all_matches) == 1:
        _, issue_path, project_dir = all_matches[0]
        issue = read_issue_from_file(issue_path)
        return IssueLookupResult(
            issue=issue, issue_path=issue_path, project_dir=project_dir
        )

    ids = ", ".join(full_id for full_id, _, _ in all_matches)
    raise IssueLookupError(f"ambiguous identifier, matches: {ids}")


def _search_directories(project_dir: Path) -> list[Path]:
    """Return issue directories to search for a given project directory."""
    dirs = [project_dir / "issues"]
    local_dir = find_project_local_directory(project_dir)
    if local_dir is not None:
        dirs.append(local_dir / "issues")
    return dirs


def resolve_issue_identifier(
    issues_dir: Path, _project_key: str, candidate: str
) -> str:
    """Resolve a full issue identifier from a user-provided value.

    Accepts a full identifier, a unique short identifier using the project key,
    or a project-context short identifier without the project key.

    :param issues_dir: Issues directory.
    :type issues_dir: Path
    :param project_key: Project key prefix.
    :type project_key: str
    :param candidate: Candidate identifier (full or short).
    :type candidate: str
    :return: Full issue identifier.
    :rtype: str
    :raises IssueLookupError: If no match or ambiguous short id.
    """
    issue_path = issues_dir / f"{candidate}.json"
    if issue_path.exists():
        return candidate

    matches = [
        identifier
        for identifier in list_issue_identifiers(issues_dir)
        if matches_issue_identifier(candidate, identifier)
    ]

    if len(matches) == 1:
        return matches[0]
    if not matches:
        raise IssueLookupError("not found")
    raise IssueLookupError("ambiguous short id")


def _find_matching_issues(issues_dir: Path, identifier: str) -> list[tuple[str, Path]]:
    matches: list[tuple[str, Path]] = []
    for full_id in list_issue_identifiers(issues_dir):
        if matches_issue_identifier(identifier, full_id):
            matches.append((full_id, issues_dir / f"{full_id}.json"))
    return matches


def _overlay_config_for_project(
    project_dir: Path, root_configuration: ProjectConfiguration | None
) -> OverlayConfig:
    if root_configuration is not None:
        return root_configuration.overlay
    config_path = project_dir.parent / ".kanbus.yml"
    if not config_path.is_file():
        return OverlayConfig(enabled=False)
    try:
        return load_project_configuration(config_path).overlay
    except (ConfigurationError, ProjectMarkerError, RuntimeError):
        return OverlayConfig(enabled=False)



from __future__ import annotations

import json
from pathlib import Path
from types import SimpleNamespace

import pytest

from kanbus import issue_lookup
from kanbus.models import OverlayConfig

from test_helpers import build_issue, build_project_configuration


def write_issue(path: Path, issue_id: str) -> None:
    issue = build_issue(issue_id)
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(
        json.dumps(issue.model_dump(by_alias=True, mode="json")),
        encoding="utf-8",
    )


def test_search_and_matching_helpers(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    project_dir = tmp_path / "project"
    local_dir = tmp_path / "project-local"
    monkeypatch.setattr(
        issue_lookup, "find_project_local_directory", lambda _p: local_dir
    )
    dirs = issue_lookup._search_directories(project_dir)
    assert dirs == [project_dir / "issues", local_dir / "issues"]

    monkeypatch.setattr(
        issue_lookup,
        "list_issue_identifiers",
        lambda _d: ["kanbus-abc123", "kanbus-def456"],
    )
    matches = issue_lookup._find_matching_issues(project_dir / "issues", "kanbus-abc")
    assert matches == [("kanbus-abc123", project_dir / "issues" / "kanbus-abc123.json")]


def test_resolve_issue_identifier_exact_unique_missing_and_ambiguous(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    issues_dir = tmp_path / "issues"
    write_issue(issues_dir / "kanbus-111aaa.json", "kanbus-111aaa")

    assert (
        issue_lookup.resolve_issue_identifier(issues_dir, "kanbus", "kanbus-111aaa")
        == "kanbus-111aaa"
    )

    monkeypatch.setattr(
        issue_lookup,
        "list_issue_identifiers",
        lambda _d: ["kanbus-111aaa", "kanbus-222bbb"],
    )
    assert (
        issue_lookup.resolve_issue_identifier(issues_dir, "kanbus", "kanbus-111")
        == "kanbus-111aaa"
    )

    monkeypatch.setattr(issue_lookup, "list_issue_identifiers", lambda _d: [])
    with pytest.raises(issue_lookup.IssueLookupError, match="not found"):
        issue_lookup.resolve_issue_identifier(issues_dir, "kanbus", "kanbus-zzz")

    monkeypatch.setattr(
        issue_lookup,
        "list_issue_identifiers",
        lambda _d: ["kanbus-111aaa", "kanbus-111bbb"],
    )
    with pytest.raises(issue_lookup.IssueLookupError, match="ambiguous short id"):
        issue_lookup.resolve_issue_identifier(issues_dir, "kanbus", "kanbus-111")


def test_overlay_config_for_project_with_root_configuration() -> None:
    config = build_project_configuration()
    overlay = OverlayConfig(enabled=True, ttl_s=123)
    config = config.model_copy(update={"overlay": overlay})
    result = issue_lookup._overlay_config_for_project(Path("/tmp/project"), config)
    assert result.enabled is True
    assert result.ttl_s == 123


def test_overlay_config_for_project_without_root_config(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    project_dir = tmp_path / "project"
    project_dir.mkdir(parents=True, exist_ok=True)
    # no .kanbus.yml => disabled
    result = issue_lookup._overlay_config_for_project(project_dir, None)
    assert result.enabled is False

    config_path = tmp_path / ".kanbus.yml"
    config_path.write_text("placeholder", encoding="utf-8")
    monkeypatch.setattr(
        issue_lookup,
        "load_project_configuration",
        lambda _p: build_project_configuration().model_copy(
            update={"overlay": OverlayConfig(enabled=True, ttl_s=77)}
        ),
    )
    result = issue_lookup._overlay_config_for_project(project_dir, None)
    assert result.enabled is True
    assert result.ttl_s == 77

    monkeypatch.setattr(
        issue_lookup,
        "load_project_configuration",
        lambda _p: (_ for _ in ()).throw(RuntimeError("bad config")),
    )
    fallback = issue_lookup._overlay_config_for_project(project_dir, None)
    assert fallback.enabled is False


def test_load_issue_from_project_exact_match_and_overlay_virtual(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root = tmp_path
    project_dir = root / "project"
    write_issue(project_dir / "issues" / "kanbus-1.json", "kanbus-1")
    monkeypatch.setattr(
        issue_lookup, "discover_project_directories", lambda _r: [project_dir]
    )
    monkeypatch.setattr(
        issue_lookup, "get_configuration_path", lambda _r: root / ".kanbus.yml"
    )
    monkeypatch.setattr(
        issue_lookup,
        "load_project_configuration",
        lambda _p: build_project_configuration(),
    )
    monkeypatch.setattr(
        issue_lookup,
        "resolve_labeled_projects",
        lambda _r: [SimpleNamespace(project_dir=project_dir, label="alpha")],
    )
    monkeypatch.setattr(issue_lookup, "load_overlay_issue", lambda *_args: None)
    monkeypatch.setattr(issue_lookup, "load_tombstone", lambda *_args: None)
    monkeypatch.setattr(
        issue_lookup,
        "resolve_issue_with_overlay",
        lambda _p, issue, *_args, **_kwargs: issue,
    )

    result = issue_lookup.load_issue_from_project(root, "kanbus-1")
    assert result.issue.identifier == "kanbus-1"
    assert result.project_dir == project_dir

    # overlay-only issue path when base issue does not exist
    overlay_issue = build_issue("kanbus-overlay")
    monkeypatch.setattr(
        issue_lookup,
        "load_overlay_issue",
        lambda *_args: SimpleNamespace(issue=overlay_issue),
    )
    monkeypatch.setattr(
        issue_lookup,
        "resolve_issue_with_overlay",
        lambda _p, _base, _overlay, *_args, **_kwargs: overlay_issue,
    )
    result = issue_lookup.load_issue_from_project(root, "kanbus-overlay")
    assert result.issue.identifier == "kanbus-overlay"
    assert result.issue_path.name == "kanbus-overlay.json"
    assert ".overlay" in str(result.issue_path)

    # exact shared file exists but overlay resolution suppresses it (e.g. tombstone)
    write_issue(project_dir / "issues" / "kanbus-suppressed.json", "kanbus-suppressed")
    monkeypatch.setattr(issue_lookup, "load_overlay_issue", lambda *_args: None)
    monkeypatch.setattr(issue_lookup, "load_tombstone", lambda *_args: object())
    monkeypatch.setattr(
        issue_lookup,
        "resolve_issue_with_overlay",
        lambda _p, issue, _overlay, tombstone, *_args, **_kwargs: (
            None if tombstone else issue
        ),
    )
    monkeypatch.setattr(issue_lookup, "list_issue_identifiers", lambda _d: [])
    with pytest.raises(issue_lookup.IssueLookupError, match="not found"):
        issue_lookup.load_issue_from_project(root, "kanbus-suppressed")


def test_load_issue_from_project_not_found_ambiguous_and_single_partial(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root = tmp_path
    project_dir = root / "project"
    issues_dir = project_dir / "issues"
    write_issue(issues_dir / "kanbus-abc111.json", "kanbus-abc111")
    write_issue(issues_dir / "kanbus-abc222.json", "kanbus-abc222")
    monkeypatch.setattr(
        issue_lookup, "discover_project_directories", lambda _r: [project_dir]
    )
    monkeypatch.setattr(
        issue_lookup,
        "load_project_configuration",
        lambda _p: (_ for _ in ()).throw(issue_lookup.ConfigurationError("bad config")),
    )
    monkeypatch.setattr(
        issue_lookup, "get_configuration_path", lambda _r: root / ".kanbus.yml"
    )
    monkeypatch.setattr(issue_lookup, "load_overlay_issue", lambda *_args: None)
    monkeypatch.setattr(issue_lookup, "load_tombstone", lambda *_args: None)

    with pytest.raises(issue_lookup.IssueLookupError, match="ambiguous identifier"):
        issue_lookup.load_issue_from_project(root, "kanbus-abc")

    write_issue(issues_dir / "kanbus-only123.json", "kanbus-only123")
    monkeypatch.setattr(
        issue_lookup,
        "list_issue_identifiers",
        lambda _d: ["kanbus-only123"],
    )
    single = issue_lookup.load_issue_from_project(root, "kanbus-only")
    assert single.issue.identifier == "kanbus-only123"

    monkeypatch.setattr(issue_lookup, "list_issue_identifiers", lambda _d: [])
    with pytest.raises(issue_lookup.IssueLookupError, match="not found"):
        issue_lookup.load_issue_from_project(root, "missing")


def test_load_issue_from_project_project_errors(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(
        issue_lookup,
        "discover_project_directories",
        lambda _r: (_ for _ in ()).throw(issue_lookup.ProjectMarkerError("no marker")),
    )
    with pytest.raises(issue_lookup.IssueLookupError, match="no marker"):
        issue_lookup.load_issue_from_project(tmp_path, "kanbus-1")

    monkeypatch.setattr(issue_lookup, "discover_project_directories", lambda _r: [])
    with pytest.raises(issue_lookup.IssueLookupError, match="project not initialized"):
        issue_lookup.load_issue_from_project(tmp_path, "kanbus-1")



from __future__ import annotations

from datetime import datetime, timedelta, timezone
from pathlib import Path
import subprocess
from types import SimpleNamespace

from kanbus import snyk_sync
from kanbus.issue_files import read_issue_from_file, write_issue_to_file
from kanbus.models import SnykConfiguration
from test_helpers import build_issue


def test_severity_to_priority_mapping() -> None:
    assert snyk_sync._severity_to_priority("critical") == 0
    assert snyk_sync._severity_to_priority("high") == 1
    assert snyk_sync._severity_to_priority("medium") == 2
    assert snyk_sync._severity_to_priority("low") == 3


def test_vuln_title_for_dependency_and_code() -> None:
    dependency_issue = {
        "attributes": {
            "key": "SNYK-001",
            "type": "package_vulnerability",
            "coordinates": [
                {
                    "representations": [
                        {
                            "dependency": {
                                "package_name": "requests",
                                "package_version": "2",
                            }
                        }
                    ]
                }
            ],
        }
    }
    code_issue = {
        "attributes": {
            "key": "SNYK-CODE-001",
            "type": "code",
            "title": "Unsanitized input",
        }
    }

    assert snyk_sync._vuln_title(dependency_issue) == "[Snyk] SNYK-001 in requests"
    assert snyk_sync._vuln_title(code_issue) == "[Snyk Code] Unsanitized input"


def test_vuln_title_dependency_without_package_name() -> None:
    dependency_issue = {
        "attributes": {
            "key": "SNYK-EMPTY",
            "type": "package_vulnerability",
            "coordinates": [{"representations": [{"dependency": {}}]}],
        }
    }
    assert snyk_sync._vuln_title(dependency_issue) == "[Snyk] SNYK-EMPTY"


def test_extract_source_location_handles_region_and_line_column() -> None:
    issue_with_region = {
        "attributes": {
            "coordinates": [
                {
                    "representations": [
                        {
                            "source_location": {
                                "file": "src/main.py",
                                "region": {
                                    "start": {"line": 10, "column": 2},
                                    "end": {"line": 12, "column": 4},
                                },
                            }
                        }
                    ]
                }
            ]
        }
    }
    issue_with_line = {
        "attributes": {
            "coordinates": [
                {
                    "representations": [
                        {
                            "sourceLocation": {
                                "file": "src/legacy.py",
                                "line": 7,
                                "col": 9,
                            }
                        }
                    ]
                }
            ]
        }
    }

    assert snyk_sync._extract_source_location(issue_with_region) == {
        "file": "src/main.py",
        "line": 10,
        "column": 2,
        "end_line": 12,
        "end_column": 4,
    }
    assert snyk_sync._extract_source_location(issue_with_line) == {
        "file": "src/legacy.py",
        "line": 7,
        "column": 9,
    }


def test_build_snippet_renders_context_window(tmp_path: Path) -> None:
    source_file = tmp_path / "src" / "app.py"
    source_file.parent.mkdir(parents=True)
    source_file.write_text(
        "\n".join(
            [
                "line 1",
                "line 2",
                "line 3",
                "line 4",
                "line 5",
            ]
        ),
        encoding="utf-8",
    )

    snippet = snyk_sync._build_snippet(tmp_path, "src/app.py", 3, 3)

    assert "### Snippet (src/app.py:1-5)" in snippet
    assert "   3 | line 3" in snippet


def test_build_snippet_returns_empty_for_empty_file(tmp_path: Path) -> None:
    source_file = tmp_path / "src" / "empty.py"
    source_file.parent.mkdir(parents=True)
    source_file.write_text("", encoding="utf-8")
    assert snyk_sync._build_snippet(tmp_path, "src/empty.py", 1, 1) == ""


def test_map_snyk_to_kanbus_sets_core_custom_fields(tmp_path: Path) -> None:
    issue = {
        "attributes": {
            "key": "SNYK-002",
            "effective_severity_level": "high",
            "type": "package_vulnerability",
            "coordinates": [
                {
                    "is_upgradeable": True,
                    "representations": [
                        {
                            "dependency": {
                                "package_name": "urllib3",
                                "package_version": "1.0.0",
                            }
                        }
                    ],
                }
            ],
            "problems": [{"source": "NVD", "id": "CVE-2026-0001"}],
        }
    }

    mapped = snyk_sync._map_snyk_to_kanbus(
        issue,
        parent_task_id="kanbus-parent",
        target_file="requirements.txt",
        repo_root=tmp_path,
    )

    assert mapped.issue_type == "sub-task"
    assert mapped.priority == 1
    assert mapped.parent == "kanbus-parent"
    assert mapped.custom["snyk_key"] == "SNYK-002"
    assert mapped.custom["snyk_severity"] == "high"


def test_extract_classes_handles_string_and_object_shapes() -> None:
    issue = {
        "attributes": {
            "classes": [
                "OWASP-A1",
                {"source": "CWE", "id": "79"},
                {"source": "CAPEC", "id": "1"},
                {"id": "missing-source"},
            ]
        }
    }
    assert snyk_sync._extract_classes(issue) == [
        "OWASP-A1",
        "CWE-79",
        "CAPEC-1",
    ]


def test_detect_repo_from_git_normalizes_github_remote(tmp_path: Path) -> None:
    subprocess.run(["git", "init"], cwd=tmp_path, check=True, capture_output=True)
    subprocess.run(
        ["git", "remote", "add", "origin", "git@github.com:AnthusAI/Kanbus.git"],
        cwd=tmp_path,
        check=True,
        capture_output=True,
    )
    assert snyk_sync._detect_repo_from_git(tmp_path) == "AnthusAI/Kanbus"


def test_detect_repo_from_git_returns_none_for_non_github_remote(
    tmp_path: Path,
) -> None:
    subprocess.run(["git", "init"], cwd=tmp_path, check=True, capture_output=True)
    subprocess.run(
        ["git", "remote", "add", "origin", "ssh://gitlab.example.com/team/repo.git"],
        cwd=tmp_path,
        check=True,
        capture_output=True,
    )
    assert snyk_sync._detect_repo_from_git(tmp_path) is None


class _FakeResponse:
    def __init__(
        self, ok: bool, status_code: int, payload: dict, text: str = ""
    ) -> None:
        self.ok = ok
        self.status_code = status_code
        self._payload = payload
        self.text = text

    def json(self) -> dict:
        return self._payload


def test_fetch_snyk_projects_handles_pagination_and_repo_filter(monkeypatch) -> None:
    calls: list[str] = []

    def fake_get(url: str, headers: dict, timeout: int) -> _FakeResponse:
        calls.append(url)
        if "limit=100" in url:
            return _FakeResponse(
                ok=True,
                status_code=200,
                payload={
                    "data": [
                        {
                            "id": "p1",
                            "attributes": {
                                "name": "AnthusAI/Kanbus:requirements.txt",
                                "target_file": "requirements.txt",
                            },
                        },
                        {
                            "id": "skip",
                            "attributes": {"name": "Other/Repo:package.json"},
                        },
                    ],
                    "links": {"next": "/rest/orgs/org/projects?page=2"},
                },
            )
        return _FakeResponse(
            ok=True,
            status_code=200,
            payload={
                "data": [
                    {
                        "id": "p2",
                        "attributes": {
                            "name": "AnthusAI/Kanbus",
                            "target_file": "repo-root",
                        },
                    }
                ],
                "links": {},
            },
        )

    monkeypatch.setattr(snyk_sync.requests, "get", fake_get)
    project_map = snyk_sync._fetch_snyk_projects(
        org_id="org", token="token", repo_filter="AnthusAI/Kanbus"
    )
    assert project_map == {"p1": "requirements.txt", "p2": "repo-root"}
    assert len(calls) == 2


def test_fetch_snyk_projects_uses_name_when_target_file_missing(monkeypatch) -> None:
    def fake_get(url: str, headers: dict, timeout: int) -> _FakeResponse:
        return _FakeResponse(
            ok=True,
            status_code=200,
            payload={
                "data": [
                    {
                        "id": "p3",
                        "attributes": {"name": "AnthusAI/Kanbus:pyproject.toml"},
                    }
                ],
                "links": {},
            },
        )

    monkeypatch.setattr(snyk_sync.requests, "get", fake_get)
    project_map = snyk_sync._fetch_snyk_projects(
        org_id="org", token="token", repo_filter="AnthusAI/Kanbus"
    )
    assert project_map == {"p3": "AnthusAI/Kanbus:pyproject.toml"}


def test_fetch_snyk_projects_raises_on_non_ok_response(monkeypatch) -> None:
    def fake_get(url: str, headers: dict, timeout: int) -> _FakeResponse:
        return _FakeResponse(
            ok=False,
            status_code=403,
            payload={},
            text="forbidden",
        )

    monkeypatch.setattr(snyk_sync.requests, "get", fake_get)
    try:
        snyk_sync._fetch_snyk_projects(org_id="org", token="token", repo_filter=None)
    except snyk_sync.SnykSyncError as error:
        assert "Snyk projects API returned 403" in str(error)
    else:
        raise AssertionError("expected SnykSyncError")


def test_fetch_snyk_issues_for_type_applies_severity_threshold(monkeypatch) -> None:
    def fake_get(url: str, headers: dict, timeout: int) -> _FakeResponse:
        return _FakeResponse(
            ok=True,
            status_code=200,
            payload={
                "data": [
                    {
                        "attributes": {
                            "key": "S-1",
                            "effective_severity_level": "critical",
                        }
                    },
                    {"attributes": {"key": "S-2", "effective_severity_level": "low"}},
                    {"attributes": {"effective_severity_level": "high"}},
                ],
                "links": {},
            },
        )

    monkeypatch.setattr(snyk_sync.requests, "get", fake_get)
    issues = snyk_sync._fetch_snyk_issues_for_type(
        org_id="org", token="token", min_priority=1, issue_type="code"
    )
    assert [item["attributes"]["key"] for item in issues] == ["S-1"]


def test_fetch_snyk_issues_for_type_raises_on_non_ok(monkeypatch) -> None:
    def fake_get(url: str, headers: dict, timeout: int) -> _FakeResponse:
        return _FakeResponse(ok=False, status_code=502, payload={}, text="gateway")

    monkeypatch.setattr(snyk_sync.requests, "get", fake_get)
    try:
        snyk_sync._fetch_snyk_issues_for_type(
            org_id="org", token="token", min_priority=1, issue_type="code"
        )
    except snyk_sync.SnykSyncError as error:
        assert "Snyk API returned 502" in str(error)
    else:
        raise AssertionError("expected SnykSyncError")


def test_fetch_snyk_issues_for_type_handles_pagination(monkeypatch) -> None:
    calls: list[str] = []

    def fake_get(url: str, headers: dict, timeout: int) -> _FakeResponse:
        calls.append(url)
        if len(calls) == 1:
            return _FakeResponse(
                ok=True,
                status_code=200,
                payload={
                    "data": [
                        {
                            "attributes": {
                                "key": "S-1",
                                "effective_severity_level": "high",
                            }
                        }
                    ],
                    "links": {"next": "/rest/orgs/org/issues?page=2"},
                },
            )
        return _FakeResponse(
            ok=True,
            status_code=200,
            payload={
                "data": [
                    {
                        "attributes": {
                            "key": "S-2",
                            "effective_severity_level": "medium",
                        }
                    }
                ],
                "links": {},
            },
        )

    monkeypatch.setattr(snyk_sync.requests, "get", fake_get)
    issues = snyk_sync._fetch_snyk_issues_for_type(
        org_id="org", token="token", min_priority=2, issue_type="code"
    )
    assert [item["attributes"]["key"] for item in issues] == ["S-1", "S-2"]
    assert len(calls) == 2


def test_fetch_all_snyk_issues_continues_when_non_package_type_fails(
    monkeypatch,
) -> None:
    def fake_fetch(org_id: str, token: str, min_priority: int, issue_type: str):
        if issue_type == "package_vulnerability":
            return [{"attributes": {"key": "S-OK"}}]
        raise snyk_sync.SnykSyncError("code endpoint failed")

    monkeypatch.setattr(snyk_sync, "_fetch_snyk_issues_for_type", fake_fetch)
    issues = snyk_sync._fetch_all_snyk_issues(
        org_id="org",
        token="token",
        min_priority=2,
        issue_types=["package_vulnerability", "code"],
    )
    assert [item["attributes"]["key"] for item in issues] == ["S-OK"]


def test_fetch_all_snyk_issues_raises_when_package_type_fails(monkeypatch) -> None:
    def fake_fetch(org_id: str, token: str, min_priority: int, issue_type: str):
        raise snyk_sync.SnykSyncError(f"{issue_type} failed")

    monkeypatch.setattr(snyk_sync, "_fetch_snyk_issues_for_type", fake_fetch)
    try:
        snyk_sync._fetch_all_snyk_issues(
            org_id="org",
            token="token",
            min_priority=2,
            issue_types=["package_vulnerability", "code"],
        )
    except snyk_sync.SnykSyncError as error:
        assert "package_vulnerability failed" in str(error)
    else:
        raise AssertionError("expected SnykSyncError")


def test_fetch_v1_enrichment_skips_non_ok_and_uses_issue_id(monkeypatch) -> None:
    def fake_post(url: str, headers: dict, json: dict, timeout: int) -> _FakeResponse:
        if "project-a" in url:
            return _FakeResponse(ok=False, status_code=500, payload={}, text="fail")
        return _FakeResponse(
            ok=True,
            status_code=200,
            payload={"issues": [{"issueData": {"id": "SNYK-1"}, "fixInfo": {}}]},
        )

    monkeypatch.setattr(snyk_sync.requests, "post", fake_post)
    enriched = snyk_sync._fetch_v1_enrichment(
        org_id="org",
        token="token",
        project_ids=["project-a", "project-b"],
    )
    assert "SNYK-1" in enriched


def test_fetch_v1_enrichment_continues_when_request_raises(monkeypatch) -> None:
    calls = {"count": 0}

    def fake_post(url: str, headers: dict, json: dict, timeout: int) -> _FakeResponse:
        calls["count"] += 1
        if calls["count"] == 1:
            raise RuntimeError("network failure")
        return _FakeResponse(
            ok=True,
            status_code=200,
            payload={"issues": [{"issueData": {"id": "SNYK-2"}}]},
        )

    monkeypatch.setattr(snyk_sync.requests, "post", fake_post)
    enriched = snyk_sync._fetch_v1_enrichment(
        org_id="org",
        token="token",
        project_ids=["project-a", "project-b"],
    )
    assert enriched == {"SNYK-2": {"issueData": {"id": "SNYK-2"}}}


def test_map_snyk_to_kanbus_code_includes_location_custom_fields(
    tmp_path: Path,
) -> None:
    source_file = tmp_path / "src" / "vuln.py"
    source_file.parent.mkdir(parents=True)
    source_file.write_text("a\nb\nc\nd\ne\n", encoding="utf-8")

    issue = {
        "attributes": {
            "key": "SNYK-CODE-42",
            "type": "code",
            "title": "Unsafely parsed input",
            "description": "User-controlled input is parsed without validation.",
            "effective_severity_level": "medium",
            "coordinates": [
                {
                    "representations": [
                        {
                            "source_location": {
                                "file": "src/vuln.py",
                                "region": {
                                    "start": {"line": 2, "column": 3},
                                    "end": {"line": 3, "column": 5},
                                },
                            }
                        }
                    ]
                }
            ],
            "classes": [{"source": "CWE", "id": "20"}],
        }
    }

    mapped = snyk_sync._map_snyk_to_kanbus(
        issue,
        parent_task_id="KAN-100",
        target_file="repo",
        repo_root=tmp_path,
    )

    assert mapped.title == "[Snyk Code] Unsafely parsed input"
    assert mapped.custom["snyk_file"] == "src/vuln.py"
    assert mapped.custom["snyk_line"] == 2
    assert mapped.custom["snyk_column"] == 3
    assert mapped.custom["snyk_end_line"] == 3
    assert mapped.custom["snyk_end_column"] == 5
    assert "**Classes:** CWE-20" in mapped.description
    assert "### Snippet (src/vuln.py:1-5)" in mapped.description


def test_map_snyk_to_kanbus_dependency_uses_v1_fix_and_meta() -> None:
    issue = {
        "attributes": {
            "key": "SNYK-DEP-99",
            "type": "package_vulnerability",
            "effective_severity_level": "high",
            "coordinates": [
                {
                    "is_upgradeable": True,
                    "representations": [
                        {
                            "dependency": {
                                "package_name": "openssl",
                                "package_version": "1.0.2",
                            }
                        }
                    ],
                }
            ],
            "problems": [
                {"source": "NVD", "id": "CVE-2026-1111"},
                {"source": "NVD", "id": "CVE-2026-2222"},
            ],
        }
    }
    v1 = {
        "issueData": {
            "title": "OpenSSL out-of-bounds read",
            "description": "Detailed advisory text",
            "cvssScore": 9.7,
            "exploitMaturity": "proof-of-concept",
        },
        "priorityScore": 860,
        "fixInfo": {"fixedIn": ["1.1.1u"]},
    }

    mapped = snyk_sync._map_snyk_to_kanbus(
        issue,
        parent_task_id="KAN-200",
        v1=v1,
        target_file="requirements.txt",
    )

    assert mapped.title == "[Snyk] SNYK-DEP-99 in openssl@1.0.2"
    assert "## OpenSSL out-of-bounds read" in mapped.description
    assert "**CVSS Score:** 9.7" in mapped.description
    assert "**Exploit Maturity:** proof-of-concept" in mapped.description
    assert "**Snyk Priority Score:** 860/1000" in mapped.description
    assert "Upgrade `openssl` to version 1.1.1u or later." in mapped.description
    assert "CVE-2026-1111" in mapped.description
    assert "CVE-2026-2222" in mapped.description


def _write_issue(issues_dir: Path, issue) -> None:
    write_issue_to_file(issue, issues_dir / f"{issue.identifier}.json")


def test_pull_from_snyk_requires_token(tmp_path: Path, monkeypatch) -> None:
    monkeypatch.delenv("SNYK_TOKEN", raising=False)
    config = SnykConfiguration.model_validate({"org_id": "org"})
    try:
        snyk_sync.pull_from_snyk(tmp_path, config, "kanbus")
    except snyk_sync.SnykSyncError as error:
        assert "SNYK_TOKEN environment variable is not set" in str(error)
    else:
        raise AssertionError("expected SnykSyncError")


def test_resolve_snyk_epics_returns_empty_when_no_categories(tmp_path: Path) -> None:
    issues_dir = tmp_path / "issues"
    issues_dir.mkdir()
    epics = snyk_sync._resolve_snyk_epics(
        issues_dir=issues_dir,
        project_key="kanbus",
        configured_id=None,
        dry_run=False,
        all_existing=set(),
        include_dependency=False,
        include_code=False,
        dependency_priority=None,
        code_priority=None,
    )
    assert epics == {}


def test_find_existing_snyk_initiative_prefers_most_recent(tmp_path: Path) -> None:
    issues_dir = tmp_path / "issues"
    issues_dir.mkdir()
    older = build_issue(
        "kanbus-1",
        issue_type="initiative",
        title=snyk_sync.SNYK_INITIATIVE_TITLE,
        labels=["snyk"],
    )
    newer = build_issue(
        "kanbus-2",
        issue_type="initiative",
        title=snyk_sync.SNYK_INITIATIVE_TITLE,
        labels=["snyk"],
    )
    older.updated_at = datetime(2026, 3, 1, tzinfo=timezone.utc)
    newer.updated_at = datetime(2026, 3, 2, tzinfo=timezone.utc)
    _write_issue(issues_dir, older)
    _write_issue(issues_dir, newer)

    found = snyk_sync._find_existing_snyk_initiative(
        issues_dir, {"kanbus-1", "kanbus-2"}
    )
    assert found == "kanbus-2"


def test_resolve_snyk_epic_updates_existing_priority(tmp_path: Path) -> None:
    issues_dir = tmp_path / "issues"
    issues_dir.mkdir()
    existing = build_issue(
        "kanbus-epic",
        issue_type="epic",
        title=snyk_sync.SNYK_DEP_EPIC_TITLE,
        parent="kanbus-init",
        priority=3,
        labels=["snyk", "security", "dependency"],
    )
    _write_issue(issues_dir, existing)
    all_existing = {"kanbus-epic"}

    epic_id = snyk_sync._resolve_snyk_epic(
        issues_dir=issues_dir,
        project_key="kanbus",
        parent_initiative="kanbus-init",
        title=snyk_sync.SNYK_DEP_EPIC_TITLE,
        category="dependency",
        priority=1,
        dry_run=False,
        all_existing=all_existing,
    )
    assert epic_id == "kanbus-epic"
    updated = read_issue_from_file(issues_dir / "kanbus-epic.json")
    assert updated.priority == 1


def test_resolve_file_task_updates_existing_task(tmp_path: Path) -> None:
    issues_dir = tmp_path / "issues"
    issues_dir.mkdir()
    task = build_issue(
        "kanbus-task",
        issue_type="task",
        title="requirements.txt",
        parent="kanbus-old-epic",
        priority=3,
        labels=[],
        custom={"snyk_target_file": "requirements.txt"},
    )
    task.updated_at = datetime.now(timezone.utc) - timedelta(days=2)
    _write_issue(issues_dir, task)

    task_id = snyk_sync._resolve_file_task(
        issues_dir=issues_dir,
        project_key="kanbus",
        target_file="requirements.txt",
        category="dependency",
        ctx=snyk_sync.FileTaskContext(
            epic_id="kanbus-new-epic",
            priority=1,
            dry_run=False,
        ),
        file_task_index={("dependency", "requirements.txt"): "kanbus-task"},
        all_existing={"kanbus-task"},
    )

    assert task_id == "kanbus-task"
    updated = read_issue_from_file(issues_dir / "kanbus-task.json")
    assert updated.parent == "kanbus-new-epic"
    assert updated.priority == 1
    assert "snyk" in updated.labels
    assert "security" in updated.labels
    assert updated.custom["snyk_target_file"] == "requirements.txt"
    assert updated.custom["snyk_category"] == "dependency"


def test_resolve_parent_epic_prefers_configured_existing_id(tmp_path: Path) -> None:
    issues_dir = tmp_path / "issues"
    issues_dir.mkdir()
    configured = build_issue("kanbus-configured", issue_type="epic", labels=["snyk"])
    _write_issue(issues_dir, configured)

    resolved = snyk_sync._resolve_parent_epic(
        issues_dir=issues_dir,
        project_key="kanbus",
        configured_id="kanbus-configured",
        dry_run=False,
        all_existing={"kanbus-configured"},
    )
    assert resolved == "kanbus-configured"


def test_resolve_parent_epic_reuses_existing_snyk_epic(tmp_path: Path) -> None:
    issues_dir = tmp_path / "issues"
    issues_dir.mkdir()
    existing = build_issue(
        "kanbus-existing",
        issue_type="epic",
        title="Snyk Vulnerabilities",
        labels=["security", "snyk"],
    )
    _write_issue(issues_dir, existing)

    resolved = snyk_sync._resolve_parent_epic(
        issues_dir=issues_dir,
        project_key="kanbus",
        configured_id=None,
        dry_run=False,
        all_existing={"kanbus-existing"},
    )
    assert resolved == "kanbus-existing"


def test_build_snyk_key_index_and_file_task_index_skip_invalid_issue_files(
    tmp_path: Path,
) -> None:
    issues_dir = tmp_path / "issues"
    issues_dir.mkdir()
    with_key = build_issue(
        "kanbus-key",
        issue_type="sub-task",
        custom={"snyk_key": "SNYK-INDEX-1"},
    )
    with_task = build_issue(
        "kanbus-task",
        issue_type="task",
        custom={"snyk_target_file": "requirements.txt", "snyk_category": "dependency"},
    )
    _write_issue(issues_dir, with_key)
    _write_issue(issues_dir, with_task)
    (issues_dir / "broken.json").write_text("{", encoding="utf-8")

    key_index = snyk_sync._build_snyk_key_index(
        {"kanbus-key", "kanbus-task", "broken"}, issues_dir
    )
    file_index = snyk_sync._build_file_task_index(
        {"kanbus-key", "kanbus-task", "broken"}, issues_dir
    )

    assert key_index["SNYK-INDEX-1"] == "kanbus-key"
    assert file_index[("dependency", "requirements.txt")] == "kanbus-task"


def test_resolve_snyk_epics_with_configured_parent_maps_requested_categories(
    tmp_path: Path, monkeypatch
) -> None:
    issues_dir = tmp_path / "issues"
    issues_dir.mkdir()
    calls: list[tuple[bool, bool]] = []
    monkeypatch.setattr(
        snyk_sync,
        "_resolve_parent_epic",
        lambda *args, **kwargs: (calls.append((True, True)) or "kanbus-parent"),
    )
    epics = snyk_sync._resolve_snyk_epics(
        issues_dir=issues_dir,
        project_key="kanbus",
        configured_id="kanbus-parent",
        dry_run=False,
        all_existing=set(),
        include_dependency=True,
        include_code=False,
        dependency_priority=1,
        code_priority=2,
    )
    assert epics == {"dependency": "kanbus-parent"}
    assert len(calls) == 1


def test_resolve_parent_epic_creates_new_epic_when_missing(
    tmp_path: Path, monkeypatch
) -> None:
    issues_dir = tmp_path / "issues"
    issues_dir.mkdir()
    monkeypatch.setattr(
        snyk_sync,
        "generate_issue_identifier",
        lambda request: SimpleNamespace(identifier="kanbus-parent-1"),
    )
    epic_id = snyk_sync._resolve_parent_epic(
        issues_dir=issues_dir,
        project_key="kanbus",
        configured_id=None,
        dry_run=False,
        all_existing=set(),
    )
    assert epic_id == "kanbus-parent-1"
    written = read_issue_from_file(issues_dir / "kanbus-parent-1.json")
    assert written.issue_type == "epic"
    assert written.title == "Snyk Vulnerabilities"


def test_resolve_snyk_epics_creates_initiative_and_both_epics(
    tmp_path: Path, monkeypatch
) -> None:
    issues_dir = tmp_path / "issues"
    issues_dir.mkdir()
    sequence = iter(
        [
            "kanbus-init-1",
            "kanbus-dep-1",
            "kanbus-code-1",
        ]
    )
    monkeypatch.setattr(
        snyk_sync,
        "generate_issue_identifier",
        lambda request: SimpleNamespace(identifier=next(sequence)),
    )
    epics = snyk_sync._resolve_snyk_epics(
        issues_dir=issues_dir,
        project_key="kanbus",
        configured_id=None,
        dry_run=False,
        all_existing=set(),
        include_dependency=True,
        include_code=True,
        dependency_priority=None,
        code_priority=None,
    )
    assert epics == {"dependency": "kanbus-dep-1", "code": "kanbus-code-1"}
    assert (
        read_issue_from_file(issues_dir / "kanbus-init-1.json").issue_type
        == "initiative"
    )
    assert (
        read_issue_from_file(issues_dir / "kanbus-dep-1.json").parent == "kanbus-init-1"
    )
    assert (
        read_issue_from_file(issues_dir / "kanbus-code-1.json").parent
        == "kanbus-init-1"
    )


def test_find_existing_snyk_initiative_and_epic_filters_non_matching(
    tmp_path: Path,
) -> None:
    issues_dir = tmp_path / "issues"
    issues_dir.mkdir()
    wrong_type = build_issue(
        "kanbus-x",
        issue_type="task",
        title=snyk_sync.SNYK_INITIATIVE_TITLE,
        labels=["snyk"],
    )
    wrong_title = build_issue(
        "kanbus-y",
        issue_type="initiative",
        title="Other title",
        labels=["snyk"],
    )
    right_init = build_issue(
        "kanbus-z",
        issue_type="initiative",
        title=snyk_sync.SNYK_INITIATIVE_TITLE,
        labels=["snyk"],
    )
    wrong_parent_epic = build_issue(
        "kanbus-epic-wrong",
        issue_type="epic",
        title=snyk_sync.SNYK_DEP_EPIC_TITLE,
        parent="other-parent",
        labels=["snyk"],
    )
    right_epic = build_issue(
        "kanbus-epic-right",
        issue_type="epic",
        title=snyk_sync.SNYK_DEP_EPIC_TITLE,
        parent="kanbus-z",
        labels=["snyk", "security"],
    )
    for issue in [wrong_type, wrong_title, right_init, wrong_parent_epic, right_epic]:
        _write_issue(issues_dir, issue)
    all_ids = {
        "kanbus-x",
        "kanbus-y",
        "kanbus-z",
        "kanbus-epic-wrong",
        "kanbus-epic-right",
    }
    assert snyk_sync._find_existing_snyk_initiative(issues_dir, all_ids) == "kanbus-z"
    assert (
        snyk_sync._find_existing_snyk_epic(
            issues_dir, all_ids, snyk_sync.SNYK_DEP_EPIC_TITLE, "kanbus-z"
        )
        == "kanbus-epic-right"
    )


def test_resolve_file_task_creates_new_task_and_dry_run_skip_write(
    tmp_path: Path, monkeypatch
) -> None:
    issues_dir = tmp_path / "issues"
    issues_dir.mkdir()
    next_id = {"value": 0}

    def _next_identifier(_request):
        next_id["value"] += 1
        return SimpleNamespace(identifier=f"kanbus-task-new-{next_id['value']}")

    monkeypatch.setattr(
        snyk_sync,
        "generate_issue_identifier",
        _next_identifier,
    )
    created = snyk_sync._resolve_file_task(
        issues_dir=issues_dir,
        project_key="kanbus",
        target_file="package.json",
        category="code",
        ctx=snyk_sync.FileTaskContext(epic_id="kanbus-epic", priority=1, dry_run=False),
        file_task_index={},
        all_existing=set(),
    )
    assert created == "kanbus-task-new-1"
    assert (
        read_issue_from_file(issues_dir / "kanbus-task-new-1.json").issue_type == "task"
    )

    dry_created = snyk_sync._resolve_file_task(
        issues_dir=issues_dir,
        project_key="kanbus",
        target_file="requirements.in",
        category="dependency",
        ctx=snyk_sync.FileTaskContext(epic_id="kanbus-epic", priority=2, dry_run=True),
        file_task_index={},
        all_existing=set(),
    )
    assert dry_created == "kanbus-task-new-2"
    assert not (issues_dir / "kanbus-task-new-2.json").exists()


def test_detect_repo_from_git_https_and_exception(tmp_path: Path, monkeypatch) -> None:
    subprocess.run(["git", "init"], cwd=tmp_path, check=True, capture_output=True)
    subprocess.run(
        ["git", "remote", "add", "origin", "https://github.com/AnthusAI/Kanbus.git"],
        cwd=tmp_path,
        check=True,
        capture_output=True,
    )
    assert snyk_sync._detect_repo_from_git(tmp_path) == "AnthusAI/Kanbus"

    monkeypatch.setattr(
        snyk_sync.subprocess,
        "run",
        lambda *args, **kwargs: (_ for _ in ()).throw(RuntimeError("boom")),
    )
    assert snyk_sync._detect_repo_from_git(tmp_path) is None


def test_build_file_task_index_defaults_missing_category_to_dependency(
    tmp_path: Path,
) -> None:
    issues_dir = tmp_path / "issues"
    issues_dir.mkdir()
    task = build_issue(
        "kanbus-task-default",
        issue_type="task",
        custom={"snyk_target_file": "pom.xml"},
    )
    _write_issue(issues_dir, task)
    index = snyk_sync._build_file_task_index({"kanbus-task-default"}, issues_dir)
    assert index[("dependency", "pom.xml")] == "kanbus-task-default"


def test_extract_source_location_and_classes_cover_skip_paths() -> None:
    no_loc_issue = {"attributes": {"coordinates": [{"representations": [{"x": 1}]}]}}
    missing_file_issue = {
        "attributes": {"coordinates": [{"representations": [{"source_location": {}}]}]}
    }
    assert snyk_sync._extract_source_location(no_loc_issue) is None
    assert snyk_sync._extract_source_location(missing_file_issue) is None

    classes_issue = {"attributes": {"classes": [{"source": "CWE"}]}}
    assert snyk_sync._extract_classes(classes_issue) == []


def test_build_snippet_handles_io_error_and_max_window(
    tmp_path: Path, monkeypatch
) -> None:
    src = tmp_path / "src" / "many.txt"
    src.parent.mkdir(parents=True)
    src.write_text("\n".join(f"line {idx}" for idx in range(1, 200)), encoding="utf-8")
    snippet = snyk_sync._build_snippet(tmp_path, "src/many.txt", 100, 160)
    assert "### Snippet (src/many.txt:" in snippet
    assert snippet.count("\n") < 60

    monkeypatch.setattr(
        Path,
        "read_text",
        lambda self, encoding=None: (_ for _ in ()).throw(OSError("denied")),
    )
    assert snyk_sync._build_snippet(tmp_path, "src/many.txt", 10, 10) == ""


def test_map_snyk_to_kanbus_covers_code_line_only_and_dependency_fix_variants() -> None:
    code_issue = {
        "attributes": {
            "key": "SNYK-CODE-LINE",
            "type": "code",
            "title": "Potential issue",
            "coordinates": [
                {
                    "representations": [
                        {"source_location": {"file": "src/app.py", "line": 10}}
                    ]
                }
            ],
        }
    }
    mapped_code = snyk_sync._map_snyk_to_kanbus(
        code_issue, parent_task_id="KAN-CODE", target_file="repo"
    )
    assert "**Location:** line 10" in mapped_code.description

    dep_issue = {
        "attributes": {
            "key": "SNYK-DEP-FIX",
            "type": "package_vulnerability",
            "effective_severity_level": "low",
            "coordinates": [
                {
                    "is_upgradeable": False,
                    "is_pinnable": True,
                    "is_fixable_snyk": False,
                    "representations": [
                        {
                            "dependency": {
                                "package_name": "pkg",
                                "package_version": "1.0",
                            }
                        }
                    ],
                }
            ],
            "problems": [],
        }
    }
    mapped_pin = snyk_sync._map_snyk_to_kanbus(dep_issue, parent_task_id="KAN-DEP")
    assert "Pin `pkg` to a patched version." in mapped_pin.description

    dep_issue["attributes"]["coordinates"][0]["is_pinnable"] = False
    dep_issue["attributes"]["coordinates"][0]["is_fixable_snyk"] = True
    mapped_fix = snyk_sync._map_snyk_to_kanbus(dep_issue, parent_task_id="KAN-DEP")
    assert "Snyk fix available" in mapped_fix.description

    dep_issue["attributes"]["coordinates"][0]["is_fixable_snyk"] = False
    mapped_manual = snyk_sync._map_snyk_to_kanbus(dep_issue, parent_task_id="KAN-DEP")
    assert "No automatic fix available" in mapped_manual.description

    dep_issue_fixed = {
        "attributes": {
            "key": "SNYK-DEP-FIXED",
            "type": "package_vulnerability",
            "effective_severity_level": "high",
            "coordinates": [
                {
                    "is_upgradeable": False,
                    "representations": [
                        {
                            "dependency": {
                                "package_name": "openssl",
                                "package_version": "1.0",
                            }
                        }
                    ],
                }
            ],
            "problems": [],
        }
    }
    mapped_fixed = snyk_sync._map_snyk_to_kanbus(
        dep_issue_fixed,
        parent_task_id="KAN-DEP",
        v1={"fixInfo": {"fixedIn": ["1.2.3"]}},
    )
    assert "Pin `openssl` to version 1.2.3 or later." in mapped_fixed.description


def test_build_snippet_returns_empty_for_missing_start_and_sets_end_from_start(
    tmp_path: Path,
) -> None:
    src = tmp_path / "src" / "short.txt"
    src.parent.mkdir(parents=True)
    src.write_text("a\nb\nc\n", encoding="utf-8")
    assert snyk_sync._build_snippet(tmp_path, "src/short.txt", None, 2) == ""
    snippet = snyk_sync._build_snippet(tmp_path, "src/short.txt", 2, 0)
    assert "src/short.txt" in snippet


def test_extract_source_location_skips_rep_with_missing_file_after_loc() -> None:
    issue = {
        "attributes": {
            "coordinates": [
                {
                    "representations": [
                        {"source_location": {"line": 7}},
                        {"source_location": {"file": "src/a.py", "line": 8}},
                    ]
                }
            ]
        }
    }
    loc = snyk_sync._extract_source_location(issue)
    assert loc is not None
    assert loc["file"] == "src/a.py"


def test_map_snyk_to_kanbus_code_location_with_column_without_end() -> None:
    issue = {
        "attributes": {
            "key": "SNYK-CODE-COL",
            "type": "code",
            "title": "Column case",
            "coordinates": [
                {
                    "representations": [
                        {
                            "source_location": {
                                "file": "src/code.py",
                                "line": 12,
                                "column": 4,
                            }
                        }
                    ]
                }
            ],
        }
    }
    mapped = snyk_sync._map_snyk_to_kanbus(issue, parent_task_id="KAN-CODE")
    assert "line 12, column 4" in mapped.description


def test_resolve_snyk_epics_configured_parent_supports_code_only(
    tmp_path: Path, monkeypatch
) -> None:
    issues_dir = tmp_path / "issues"
    issues_dir.mkdir()
    monkeypatch.setattr(
        snyk_sync,
        "_resolve_parent_epic",
        lambda *args, **kwargs: "kanbus-parent",
    )
    epics = snyk_sync._resolve_snyk_epics(
        issues_dir=issues_dir,
        project_key="kanbus",
        configured_id="kanbus-parent",
        dry_run=False,
        all_existing=set(),
        include_dependency=False,
        include_code=True,
        dependency_priority=2,
        code_priority=1,
    )
    assert epics == {"code": "kanbus-parent"}


def test_resolve_snyk_initiative_returns_existing_without_write(tmp_path: Path) -> None:
    issues_dir = tmp_path / "issues"
    issues_dir.mkdir()
    existing = build_issue(
        "kanbus-init-existing",
        issue_type="initiative",
        title=snyk_sync.SNYK_INITIATIVE_TITLE,
        labels=["snyk"],
    )
    _write_issue(issues_dir, existing)
    resolved = snyk_sync._resolve_snyk_initiative(
        issues_dir=issues_dir,
        project_key="kanbus",
        dry_run=False,
        all_existing={"kanbus-init-existing"},
    )
    assert resolved == "kanbus-init-existing"


def test_find_existing_filters_for_snyk_labels(tmp_path: Path) -> None:
    issues_dir = tmp_path / "issues"
    issues_dir.mkdir()
    initiative = build_issue(
        "kanbus-init-no-label",
        issue_type="initiative",
        title=snyk_sync.SNYK_INITIATIVE_TITLE,
        labels=[],
    )
    epic = build_issue(
        "kanbus-epic-no-label",
        issue_type="epic",
        title=snyk_sync.SNYK_DEP_EPIC_TITLE,
        parent="kanbus-parent",
        labels=[],
    )
    _write_issue(issues_dir, initiative)
    _write_issue(issues_dir, epic)
    assert (
        snyk_sync._find_existing_snyk_initiative(issues_dir, {"kanbus-init-no-label"})
        is None
    )
    assert (
        snyk_sync._find_existing_snyk_epic(
            issues_dir,
            {"kanbus-epic-no-label"},
            snyk_sync.SNYK_DEP_EPIC_TITLE,
            "kanbus-parent",
        )
        is None
    )


def test_find_existing_parent_epic_filters_non_matching_rows(tmp_path: Path) -> None:
    issues_dir = tmp_path / "issues"
    issues_dir.mkdir()
    non_epic = build_issue(
        "kanbus-task",
        issue_type="task",
        title="Snyk Vulnerabilities",
        labels=["snyk"],
    )
    wrong_title = build_issue(
        "kanbus-epic-wrong-title",
        issue_type="epic",
        title="Different title",
        labels=["snyk"],
    )
    wrong_label = build_issue(
        "kanbus-epic-wrong-label",
        issue_type="epic",
        title="Snyk Vulnerabilities",
        labels=["security"],
    )
    for issue in [non_epic, wrong_title, wrong_label]:
        _write_issue(issues_dir, issue)
    assert (
        snyk_sync._find_existing_parent_epic(
            issues_dir,
            {"kanbus-task", "kanbus-epic-wrong-title", "kanbus-epic-wrong-label"},
        )
        is None
    )


def test_pull_from_snyk_missing_issues_dir_raises(tmp_path: Path, monkeypatch) -> None:
    monkeypatch.setenv("SNYK_TOKEN", "token")
    config = SnykConfiguration.model_validate({"org_id": "org"})
    import kanbus.project as project_module

    project_dir = tmp_path / "project"
    project_dir.mkdir(parents=True)
    monkeypatch.setattr(
        project_module, "load_project_directory", lambda root: project_dir
    )
    try:
        snyk_sync.pull_from_snyk(tmp_path, config, "kanbus")
    except snyk_sync.SnykSyncError as error:
        assert "issues directory does not exist" in str(error)
    else:
        raise AssertionError("expected SnykSyncError")


def test_pull_from_snyk_updated_path_and_skip_missing_project_map(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    project_dir = tmp_path / "project"
    issues_dir = project_dir / "issues"
    issues_dir.mkdir(parents=True)
    monkeypatch.setenv("SNYK_TOKEN", "token")

    import kanbus.project as project_module

    monkeypatch.setattr(
        project_module, "load_project_directory", lambda root: project_dir
    )
    monkeypatch.setattr(
        snyk_sync, "_detect_repo_from_git", lambda root: "AnthusAI/Kanbus"
    )
    monkeypatch.setattr(
        snyk_sync,
        "_fetch_snyk_projects",
        lambda org_id, token, repo_filter: {"p1": "requirements.txt"},
    )
    monkeypatch.setattr(
        snyk_sync,
        "_fetch_all_snyk_issues",
        lambda org_id, token, min_priority, issue_types: [
            {
                "relationships": {"scan_item": {"data": {"id": "missing-project"}}},
                "attributes": {
                    "key": "SNYK-SKIP-1",
                    "type": "package_vulnerability",
                    "effective_severity_level": "low",
                },
            },
            {
                "relationships": {"scan_item": {"data": {"id": "p1"}}},
                "attributes": {
                    "key": "SNYK-UPD-1",
                    "type": "package_vulnerability",
                    "effective_severity_level": "high",
                },
            },
        ],
    )
    monkeypatch.setattr(
        snyk_sync, "_fetch_v1_enrichment", lambda org_id, token, project_ids: {}
    )
    monkeypatch.setattr(
        snyk_sync,
        "_resolve_snyk_epics",
        lambda *args, **kwargs: {"dependency": "kanbus-epic"},
    )
    monkeypatch.setattr(
        snyk_sync,
        "_build_snyk_key_index",
        lambda *_args: {"SNYK-UPD-1": "kanbus-existing-1"},
    )
    monkeypatch.setattr(snyk_sync, "_build_file_task_index", lambda *_args: {})
    monkeypatch.setattr(
        snyk_sync,
        "_resolve_file_task",
        lambda *args, **kwargs: "kanbus-task-dep",
    )
    monkeypatch.setattr(
        snyk_sync,
        "_map_snyk_to_kanbus",
        lambda vuln, task_id, v1_data, target_file, root: build_issue(
            "__placeholder__",
            title="Updated vuln",
            issue_type="sub-task",
            parent=task_id,
            custom={"snyk_key": vuln["attributes"]["key"]},
        ),
    )
    existing = build_issue("kanbus-existing-1", issue_type="sub-task")
    _write_issue(issues_dir, existing)

    def _read_issue_fail(path: Path):
        if path.name == "kanbus-existing-1.json":
            raise RuntimeError("read failed")
        return read_issue_from_file(path)

    monkeypatch.setattr(snyk_sync, "read_issue_from_file", _read_issue_fail)
    writes: list[Path] = []
    monkeypatch.setattr(
        snyk_sync,
        "write_issue_to_file",
        lambda issue, issue_path: writes.append(issue_path),
    )

    config = SnykConfiguration.model_validate({"org_id": "org", "min_severity": "low"})
    result = snyk_sync.pull_from_snyk(tmp_path, config, "kanbus", dry_run=False)
    assert result.updated == 1
    assert result.pulled == 0
    assert len(writes) == 1
    assert "Warning: failed to read existing issue file" in capsys.readouterr().out


def test_pull_from_snyk_skips_when_no_epic_for_category(
    tmp_path: Path, monkeypatch
) -> None:
    project_dir = tmp_path / "project"
    issues_dir = project_dir / "issues"
    issues_dir.mkdir(parents=True)
    monkeypatch.setenv("SNYK_TOKEN", "token")

    import kanbus.project as project_module

    monkeypatch.setattr(
        project_module, "load_project_directory", lambda root: project_dir
    )
    monkeypatch.setattr(
        snyk_sync, "_detect_repo_from_git", lambda root: "AnthusAI/Kanbus"
    )
    monkeypatch.setattr(
        snyk_sync,
        "_fetch_snyk_projects",
        lambda org_id, token, repo_filter: {"p1": "requirements.txt"},
    )
    monkeypatch.setattr(
        snyk_sync,
        "_fetch_all_snyk_issues",
        lambda org_id, token, min_priority, issue_types: [
            {
                "relationships": {"scan_item": {"data": {"id": "p1"}}},
                "attributes": {
                    "key": "SNYK-NO-EPIC",
                    "type": "package_vulnerability",
                    "effective_severity_level": "high",
                },
            }
        ],
    )
    monkeypatch.setattr(snyk_sync, "_fetch_v1_enrichment", lambda *_args: {})
    monkeypatch.setattr(snyk_sync, "_resolve_snyk_epics", lambda *args, **kwargs: {})
    monkeypatch.setattr(snyk_sync, "_build_snyk_key_index", lambda *_args: {})
    monkeypatch.setattr(snyk_sync, "_build_file_task_index", lambda *_args: {})
    called = {"resolve_task": 0, "write": 0}
    monkeypatch.setattr(
        snyk_sync,
        "_resolve_file_task",
        lambda *args, **kwargs: called.__setitem__(
            "resolve_task", called["resolve_task"] + 1
        ),
    )
    monkeypatch.setattr(
        snyk_sync,
        "write_issue_to_file",
        lambda *args, **kwargs: called.__setitem__("write", called["write"] + 1),
    )
    config = SnykConfiguration.model_validate({"org_id": "org"})
    result = snyk_sync.pull_from_snyk(tmp_path, config, "kanbus", dry_run=False)
    assert result.pulled == 0
    assert result.updated == 0
    assert called["resolve_task"] == 0
    assert called["write"] == 0


def test_pull_from_snyk_updated_path_preserves_created_at_when_existing_readable(
    tmp_path: Path, monkeypatch
) -> None:
    project_dir = tmp_path / "project"
    issues_dir = project_dir / "issues"
    issues_dir.mkdir(parents=True)
    monkeypatch.setenv("SNYK_TOKEN", "token")

    import kanbus.project as project_module

    monkeypatch.setattr(
        project_module, "load_project_directory", lambda root: project_dir
    )
    monkeypatch.setattr(
        snyk_sync, "_detect_repo_from_git", lambda root: "AnthusAI/Kanbus"
    )
    monkeypatch.setattr(
        snyk_sync,
        "_fetch_snyk_projects",
        lambda org_id, token, repo_filter: {"p1": "requirements.txt"},
    )
    monkeypatch.setattr(
        snyk_sync,
        "_fetch_all_snyk_issues",
        lambda org_id, token, min_priority, issue_types: [
            {
                "relationships": {"scan_item": {"data": {"id": "p1"}}},
                "attributes": {
                    "key": "SNYK-UPD-READ",
                    "type": "package_vulnerability",
                    "effective_severity_level": "high",
                },
            }
        ],
    )
    monkeypatch.setattr(snyk_sync, "_fetch_v1_enrichment", lambda *_args: {})
    monkeypatch.setattr(
        snyk_sync,
        "_resolve_snyk_epics",
        lambda *args, **kwargs: {"dependency": "kanbus-epic"},
    )
    monkeypatch.setattr(
        snyk_sync,
        "_build_snyk_key_index",
        lambda *_args: {"SNYK-UPD-READ": "kanbus-existing-2"},
    )
    monkeypatch.setattr(snyk_sync, "_build_file_task_index", lambda *_args: {})
    monkeypatch.setattr(
        snyk_sync,
        "_resolve_file_task",
        lambda *args, **kwargs: "kanbus-task",
    )
    monkeypatch.setattr(
        snyk_sync,
        "_map_snyk_to_kanbus",
        lambda vuln, task_id, v1_data, target_file, root: build_issue(
            "__placeholder__",
            title="Updated readable",
            issue_type="sub-task",
            parent=task_id,
            custom={"snyk_key": vuln["attributes"]["key"]},
        ),
    )
    existing = build_issue("kanbus-existing-2", issue_type="sub-task")
    existing.created_at = datetime(2024, 1, 1, tzinfo=timezone.utc)
    _write_issue(issues_dir, existing)
    written: dict[str, object] = {}

    def _capture_write(issue, issue_path):
        written["created_at"] = issue.created_at
        written["path"] = issue_path

    monkeypatch.setattr(snyk_sync, "write_issue_to_file", _capture_write)
    config = SnykConfiguration.model_validate({"org_id": "org"})
    result = snyk_sync.pull_from_snyk(tmp_path, config, "kanbus", dry_run=False)
    assert result.updated == 1
    assert written["path"].name == "kanbus-existing-2.json"
    assert written["created_at"] == datetime(2024, 1, 1, tzinfo=timezone.utc)


def test_pull_from_snyk_groups_dedups_and_writes_issues(
    tmp_path: Path, monkeypatch
) -> None:
    project_dir = tmp_path / "project"
    issues_dir = project_dir / "issues"
    issues_dir.mkdir(parents=True)
    monkeypatch.setenv("SNYK_TOKEN", "token")

    import kanbus.project as project_module

    monkeypatch.setattr(
        project_module, "load_project_directory", lambda root: project_dir
    )
    monkeypatch.setattr(
        snyk_sync, "_detect_repo_from_git", lambda root: "AnthusAI/Kanbus"
    )
    monkeypatch.setattr(
        snyk_sync,
        "_fetch_snyk_projects",
        lambda org_id, token, repo_filter: {
            "p1": "requirements.txt",
            "p2": "src/app.py",
        },
    )
    monkeypatch.setattr(
        snyk_sync,
        "_fetch_all_snyk_issues",
        lambda org_id, token, min_priority, issue_types: [
            {
                "relationships": {"scan_item": {"data": {"id": "p1"}}},
                "attributes": {
                    "key": "SNYK-DEP-1",
                    "type": "package_vulnerability",
                    "effective_severity_level": "high",
                },
            },
            {
                "relationships": {"scan_item": {"data": {"id": "p1"}}},
                "attributes": {
                    "key": "SNYK-DEP-1",
                    "type": "package_vulnerability",
                    "effective_severity_level": "high",
                },
            },
            {
                "relationships": {"scan_item": {"data": {"id": "p2"}}},
                "attributes": {
                    "key": "SNYK-CODE-1",
                    "type": "code",
                    "effective_severity_level": "medium",
                },
            },
        ],
    )
    monkeypatch.setattr(
        snyk_sync, "_fetch_v1_enrichment", lambda org_id, token, project_ids: {}
    )

    epics_calls: list[tuple[bool, bool, int | None, int | None]] = []

    def fake_resolve_epics(
        issues_dir: Path,
        project_key: str,
        configured_id: str | None,
        dry_run: bool,
        all_existing: set[str],
        include_dependency: bool,
        include_code: bool,
        dependency_priority: int | None,
        code_priority: int | None,
    ) -> dict[str, str]:
        epics_calls.append(
            (include_dependency, include_code, dependency_priority, code_priority)
        )
        return {"dependency": "kanbus-dep-epic", "code": "kanbus-code-epic"}

    monkeypatch.setattr(snyk_sync, "_resolve_snyk_epics", fake_resolve_epics)
    monkeypatch.setattr(snyk_sync, "_build_snyk_key_index", lambda *_args: {})
    monkeypatch.setattr(snyk_sync, "_build_file_task_index", lambda *_args: {})
    monkeypatch.setattr(
        snyk_sync,
        "_resolve_file_task",
        lambda issues_dir, project_key, target_file, category, ctx, file_task_index, all_existing: f"kanbus-task-{category}",
    )
    next_id = {"value": 0}

    def fake_generate_issue_identifier(request):
        next_id["value"] += 1
        return SimpleNamespace(
            identifier=f"{request.prefix}-new-{next_id['value']:03d}"
        )

    monkeypatch.setattr(
        snyk_sync, "generate_issue_identifier", fake_generate_issue_identifier
    )
    monkeypatch.setattr(
        snyk_sync,
        "_map_snyk_to_kanbus",
        lambda vuln, task_id, v1_data, target_file, root: build_issue(
            "__placeholder__",
            title=f"{target_file}:{vuln['attributes']['key']}",
            issue_type="sub-task",
            parent=task_id,
            labels=["security", "snyk"],
            custom={"snyk_key": vuln["attributes"]["key"]},
        ),
    )

    config = SnykConfiguration.model_validate(
        {"org_id": "org", "min_severity": "medium"}
    )
    result = snyk_sync.pull_from_snyk(tmp_path, config, "kanbus", dry_run=False)

    assert result.pulled == 2
    assert result.updated == 0
    assert result.skipped == 0
    assert epics_calls == [(True, True, 1, 2)]

    written_files = sorted(path.name for path in issues_dir.glob("*.json"))
    assert written_files == ["kanbus-new-001.json", "kanbus-new-002.json"]
    for file_name in written_files:
        written = read_issue_from_file(issues_dir / file_name)
        assert written.parent in {"kanbus-task-dependency", "kanbus-task-code"}



"""Rich text quality signals for CLI output.

Detects and repairs common malformations in issue descriptions and comments
submitted through the CLI, and emits actionable warnings and suggestions.
"""

from __future__ import annotations

import re

import click
from dataclasses import dataclass, field

_MARKDOWN_PATTERNS = [
    re.compile(r"^#{1,6}\s", re.MULTILINE),
    re.compile(r"\*\*[^*]+\*\*"),
    re.compile(r"\*[^*]+\*"),
    re.compile(r"__[^_]+__"),
    re.compile(r"_[^_]+_"),
    re.compile(r"^```", re.MULTILINE),
    re.compile(r"`[^`]+`"),
    re.compile(r"^>", re.MULTILINE),
    re.compile(r"^\s*[-*+]\s", re.MULTILINE),
    re.compile(r"^\s*\d+\.\s", re.MULTILINE),
    re.compile(r"^---$", re.MULTILINE),
    re.compile(r"\[.+\]\(.+\)"),
]

_DIAGRAM_FENCE_PATTERN = re.compile(
    r"^```\s*(mermaid|plantuml|d2)\s*$", re.MULTILINE | re.IGNORECASE
)

_LITERAL_NEWLINE_PATTERN = re.compile(r"\\n")

_SUPPORTED_DIAGRAM_LANGUAGES = ("mermaid", "plantuml", "d2")


@dataclass
class TextQualityResult:
    """Result of applying quality signals to a text value.

    :param text: The (possibly repaired) text after processing.
    :type text: str
    :param warnings: Warning messages describing automatic transformations.
    :type warnings: list[str]
    :param suggestions: Non-blocking suggestion messages.
    :type suggestions: list[str]
    :param escape_sequences_repaired: True if literal \\n sequences were replaced.
    :type escape_sequences_repaired: bool
    """

    text: str
    warnings: list[str] = field(default_factory=list)
    suggestions: list[str] = field(default_factory=list)
    escape_sequences_repaired: bool = False


def repair_escape_sequences(text: str) -> tuple[str, bool]:
    """Replace literal backslash-n sequences with real newline characters.

    :param text: Input text that may contain literal ``\\n`` sequences.
    :type text: str
    :return: Tuple of (repaired text, whether any replacements were made).
    :rtype: tuple[str, bool]
    """
    if _LITERAL_NEWLINE_PATTERN.search(text):
        return _LITERAL_NEWLINE_PATTERN.sub("\n", text), True
    return text, False


def has_markdown_formatting(text: str) -> bool:
    """Return True if the text contains at least one Markdown element.

    :param text: Text to inspect.
    :type text: str
    :return: True if any Markdown formatting is detected.
    :rtype: bool
    """
    return any(pattern.search(text) for pattern in _MARKDOWN_PATTERNS)


def has_diagram_block(text: str) -> bool:
    """Return True if the text contains at least one diagram fenced code block.

    :param text: Text to inspect.
    :type text: str
    :return: True if a mermaid, plantuml, or d2 block is found.
    :rtype: bool
    """
    return bool(_DIAGRAM_FENCE_PATTERN.search(text))


def apply_text_quality_signals(text: str) -> TextQualityResult:
    """Apply all quality signals to the given text.

    Repairs literal escape sequences, then checks for missing Markdown
    formatting and missing diagram blocks, collecting warnings and suggestions.

    :param text: Raw text from the CLI argument.
    :type text: str
    :return: Quality result with repaired text, warnings, and suggestions.
    :rtype: TextQualityResult
    """
    repaired_text, sequences_were_repaired = repair_escape_sequences(text)

    warnings: list[str] = []
    suggestions: list[str] = []

    if sequences_were_repaired:
        warnings.append(
            "WARNING: Literal \\n escape sequences were detected and replaced with real "
            "newlines.\n"
            "  To pass multi-line text correctly, use a heredoc or $'...\\n...' syntax:\n"
            "    kbs create \"Title\" --description $'First line\\nSecond line'\n"
            '    kbs create "Title" --description "$(cat <<\'EOF\'\\nFirst line\\n'
            'Second line\\nEOF\\n)"'
        )

    if not has_markdown_formatting(repaired_text):
        suggestions.append(
            "SUGGESTION: Markdown formatting is supported in descriptions and comments.\n"
            "  Use headings, bold, code blocks, and lists when they improve readability."
        )

    if not has_diagram_block(repaired_text):
        suggestions.append(
            "SUGGESTION: Diagrams can be embedded using fenced code blocks.\n"
            "  Supported: mermaid, plantuml, d2. Use these to visualize flows or structures."
        )

    return TextQualityResult(
        text=repaired_text,
        warnings=warnings,
        suggestions=suggestions,
        escape_sequences_repaired=sequences_were_repaired,
    )


def emit_signals(
    result: TextQualityResult,
    context: str,
    issue_id: str | None = None,
    comment_id: str | None = None,
    is_update: bool = False,
) -> None:
    """Print quality signal messages to stderr.

    Emits warnings first, then suggestions, and finally a follow-up command
    hint if any suggestions were generated.

    :param result: The quality result to emit signals from.
    :type result: TextQualityResult
    :param context: Human-readable label for what was processed (e.g. "description").
    :type context: str
    :param issue_id: Issue identifier for follow-up command hints.
    :type issue_id: str | None
    :param comment_id: Comment identifier for comment-update hints.
    :type comment_id: str | None
    :param is_update: Whether this is an update operation (affects hint wording).
    :type is_update: bool
    """
    for warning in result.warnings:
        click.echo(warning, err=True)

    if result.suggestions:
        for suggestion in result.suggestions:
            click.echo(suggestion, err=True)

        if issue_id:
            _emit_follow_up_hint(
                context=context,
                issue_id=issue_id,
                comment_id=comment_id,
                is_update=is_update,
            )


def _emit_follow_up_hint(
    context: str,
    issue_id: str,
    comment_id: str | None,
    is_update: bool,
) -> None:
    """Emit a single ready-to-run follow-up command hint to stderr.

    :param context: Human-readable label for what was processed.
    :type context: str
    :param issue_id: Issue identifier.
    :type issue_id: str
    :param comment_id: Comment identifier, if applicable.
    :type comment_id: str | None
    :param is_update: Whether this is an update operation.
    :type is_update: bool
    """
    if comment_id:
        hint = (
            f"  -> To update this comment: "
            f'kbs comment update {issue_id} {comment_id} "<your improved comment here>"'
        )
    else:
        hint = (
            f"  -> To update the {context}: "
            f'kbs update {issue_id} --description "<your improved description here>"'
        )
    click.echo(hint, err=True)



from __future__ import annotations

from pathlib import Path
from types import SimpleNamespace

from click.testing import CliRunner
import pytest

from kanbus import cli
from kanbus.console_snapshot import ConsoleSnapshotError
from kanbus.maintenance import ProjectStatsError, ProjectValidationError


def _run(args: list[str]) -> object:
    runner = CliRunner()
    return runner.invoke(cli.cli, args)


def test_console_snapshot_success_and_error(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.setattr(cli.Path, "cwd", lambda: tmp_path)
    monkeypatch.setattr(cli, "build_console_snapshot", lambda _root: {"ok": True})

    result = _run(["console", "snapshot"])
    assert result.exit_code == 0
    assert '"ok": true' in result.output

    monkeypatch.setattr(
        cli,
        "build_console_snapshot",
        lambda _root: (_ for _ in ()).throw(ConsoleSnapshotError("snap fail")),
    )
    result_error = _run(["console", "snapshot"])
    assert result_error.exit_code != 0
    assert "snap fail" in result_error.output


@pytest.mark.parametrize(
    "args",
    [
        ["console", "focus", "kanbus-1"],
        ["console", "focus", "kanbus-1", "--comment", "c1"],
        ["console", "unfocus"],
        ["console", "view", "issues"],
        ["console", "search", "text"],
        ["console", "search", "--clear"],
        ["console", "maximize"],
        ["console", "restore"],
        ["console", "close-detail"],
        ["console", "toggle-settings"],
        ["console", "reload"],
        ["console", "set-setting", "mode", "dark"],
        ["console", "collapse", "open"],
        ["console", "collapse-column", "open"],
        ["console", "expand", "open"],
        ["console", "expand-column", "open"],
        ["console", "select", "kanbus-1"],
    ],
)
def test_console_deprecated_commands(args: list[str]) -> None:
    result = _run(args)
    assert result.exit_code != 0
    assert "deprecated" in result.output.lower()


def test_console_status_and_get_commands(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.setattr(cli.Path, "cwd", lambda: tmp_path)

    monkeypatch.setattr(cli, "fetch_console_ui_state", lambda _root: None)
    result_none = _run(["console", "status"])
    assert result_none.exit_code == 0
    assert "Console server is not running." in result_none.output

    state = {
        "focused_issue_id": "kanbus-1",
        "view_mode": "issues",
        "search_query": "abc",
    }
    monkeypatch.setattr(cli, "fetch_console_ui_state", lambda _root: state)
    result_status = _run(["console", "status"])
    assert "focus:  kanbus-1" in result_status.output
    assert "view:   issues" in result_status.output
    assert "search: abc" in result_status.output

    result_focus = _run(["console", "get", "focus"])
    assert result_focus.exit_code == 0
    assert "kanbus-1" in result_focus.output

    result_view = _run(["console", "get", "view"])
    assert result_view.exit_code == 0
    assert "issues" in result_view.output

    result_search = _run(["console", "get", "search"])
    assert result_search.exit_code == 0
    assert "abc" in result_search.output

    monkeypatch.setattr(cli, "fetch_console_ui_state", lambda _root: {})
    assert "none" in _run(["console", "get", "focus"]).output
    assert "none" in _run(["console", "get", "view"]).output
    assert "none" in _run(["console", "get", "search"]).output

    monkeypatch.setattr(cli, "fetch_console_ui_state", lambda _root: None)
    assert "Console server is not running." in _run(["console", "get", "focus"]).output
    assert "Console server is not running." in _run(["console", "get", "view"]).output
    assert "Console server is not running." in _run(["console", "get", "search"]).output


def test_validate_command_success_and_error(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.setattr(cli.Path, "cwd", lambda: tmp_path)
    called: list[Path] = []
    monkeypatch.setattr(cli, "validate_project", lambda root: called.append(root))

    result = _run(["validate"])
    assert result.exit_code == 0
    assert called == [tmp_path]

    monkeypatch.setattr(
        cli,
        "validate_project",
        lambda _root: (_ for _ in ()).throw(ProjectValidationError("invalid project")),
    )
    result_error = _run(["validate"])
    assert result_error.exit_code != 0
    assert "invalid project" in result_error.output


def test_stats_command_success_and_error(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.setattr(cli.Path, "cwd", lambda: tmp_path)
    monkeypatch.setattr(
        cli,
        "collect_project_stats",
        lambda _root: SimpleNamespace(
            total=3,
            open_count=2,
            closed_count=1,
            type_counts={"task": 2, "bug": 1},
        ),
    )

    result = _run(["stats"])
    assert result.exit_code == 0
    assert "total issues: 3" in result.output
    assert "open issues: 2" in result.output
    assert "closed issues: 1" in result.output
    assert "type: bug: 1" in result.output
    assert "type: task: 2" in result.output

    monkeypatch.setattr(
        cli,
        "collect_project_stats",
        lambda _root: (_ for _ in ()).throw(ProjectStatsError("stats fail")),
    )
    result_error = _run(["stats"])
    assert result_error.exit_code != 0
    assert "stats fail" in result_error.output



from __future__ import annotations

import io
import json
from pathlib import Path

import pytest

from kanbus import daemon_server
from kanbus.daemon_protocol import ProtocolError, RequestEnvelope, ResponseEnvelope
from kanbus.index import IssueIndex

from test_helpers import build_issue


def test_daemon_core_warm_start_uses_cache_or_builds_index(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    core = daemon_server.DaemonCore(root=tmp_path)
    project_dir = tmp_path / "project"
    issues_dir = project_dir / "issues"
    cache_path = tmp_path / ".cache" / "index.json"
    monkeypatch.setattr(daemon_server, "load_project_directory", lambda _r: project_dir)
    monkeypatch.setattr(daemon_server, "get_index_cache_path", lambda _r: cache_path)

    built_index = IssueIndex(by_id={"kanbus-1": build_issue("kanbus-1")})
    monkeypatch.setattr(daemon_server, "load_cache_if_valid", lambda *_a: None)
    monkeypatch.setattr(
        daemon_server, "build_index_from_directory", lambda _d: built_index
    )
    monkeypatch.setattr(
        daemon_server, "collect_issue_file_mtimes", lambda _d: {"x": 1.0}
    )
    called: dict[str, bool] = {}
    monkeypatch.setattr(
        daemon_server, "write_cache", lambda *_a: called.setdefault("write_cache", True)
    )
    core.warm_start()
    assert core.state.index is built_index
    assert called.get("write_cache") is True

    cached_index = IssueIndex(by_id={"kanbus-cached": build_issue("kanbus-cached")})
    monkeypatch.setattr(daemon_server, "load_cache_if_valid", lambda *_a: cached_index)
    called.clear()
    core.warm_start()
    assert core.state.index is cached_index
    assert called == {}
    assert issues_dir == project_dir / "issues"


def test_daemon_core_handle_request_variants(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    core = daemon_server.DaemonCore(root=tmp_path)
    issue = build_issue("kanbus-1")
    monkeypatch.setattr(core, "_load_index", lambda: [issue])

    ping = core.handle_request(
        RequestEnvelope(
            protocol_version="1.0",
            request_id="req-ping",
            action="ping",
            payload={},
        )
    )
    assert ping.status == "ok"
    assert ping.result == {"status": "ok"}

    shutdown = core.handle_request(
        RequestEnvelope(
            protocol_version="1.0",
            request_id="req-stop",
            action="shutdown",
            payload={},
        )
    )
    assert shutdown.status == "ok"
    assert shutdown.result == {"status": "stopping"}

    listing = core.handle_request(
        RequestEnvelope(
            protocol_version="1.0",
            request_id="req-list",
            action="index.list",
            payload={},
        )
    )
    assert listing.status == "ok"
    assert listing.result is not None
    assert listing.result["issues"][0]["id"] == "kanbus-1"

    unknown = core.handle_request(
        RequestEnvelope(
            protocol_version="1.0",
            request_id="req-unknown",
            action="nope",
            payload={},
        )
    )
    assert unknown.status == "error"
    assert unknown.error is not None
    assert unknown.error.code == "unknown_action"

    with pytest.raises(ProtocolError, match="protocol version mismatch"):
        core.handle_request(
            RequestEnvelope(
                protocol_version="2.0",
                request_id="req-bad-version",
                action="ping",
                payload={},
            )
        )


def test_daemon_core_load_index_rebuild_and_cached(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    core = daemon_server.DaemonCore(root=tmp_path)
    project_dir = tmp_path / "project"
    cache_path = tmp_path / ".cache" / "index.json"
    monkeypatch.setattr(daemon_server, "load_project_directory", lambda _r: project_dir)
    monkeypatch.setattr(daemon_server, "get_index_cache_path", lambda _r: cache_path)

    built = IssueIndex(by_id={"kanbus-1": build_issue("kanbus-1")})
    monkeypatch.setattr(daemon_server, "load_cache_if_valid", lambda *_a: None)
    monkeypatch.setattr(daemon_server, "build_index_from_directory", lambda _d: built)
    monkeypatch.setattr(
        daemon_server, "collect_issue_file_mtimes", lambda _d: {"k": 1.0}
    )
    monkeypatch.setattr(daemon_server, "write_cache", lambda *_a: None)
    assert [issue.identifier for issue in core._load_index()] == ["kanbus-1"]

    cached = IssueIndex(by_id={"kanbus-cached": build_issue("kanbus-cached")})
    monkeypatch.setattr(daemon_server, "load_cache_if_valid", lambda *_a: cached)
    assert [issue.identifier for issue in core._load_index()] == ["kanbus-cached"]


def test_raw_request_and_response_error_helpers(tmp_path: Path) -> None:
    core = daemon_server.DaemonCore(tmp_path)
    response, action = daemon_server._handle_raw_request(core, b"{not-json")
    assert action is None
    assert response.status == "error"
    assert response.error is not None
    assert response.error.code == "internal_error"

    unsupported = daemon_server._build_protocol_error_response(
        "req-1", ProtocolError("protocol version unsupported")
    )
    assert unsupported.error is not None
    assert unsupported.error.code == "protocol_version_unsupported"

    mismatch = daemon_server._build_protocol_error_response(
        "req-2", ProtocolError("protocol version mismatch")
    )
    assert mismatch.error is not None
    assert mismatch.error.code == "protocol_version_mismatch"

    internal = daemon_server._build_internal_error_response(
        "req-3", RuntimeError("boom")
    )
    assert internal.error is not None
    assert internal.error.message == "boom"


def test_handle_request_for_testing_and_raw_payload_for_testing(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    request = RequestEnvelope(
        protocol_version="1.0",
        request_id="req-1",
        action="ping",
        payload={},
    )
    monkeypatch.setattr(
        daemon_server.DaemonCore,
        "handle_request",
        lambda _self, _req: ResponseEnvelope(
            protocol_version="1.0",
            request_id="req-1",
            status="ok",
            result={"status": "ok"},
        ),
    )
    response = daemon_server.handle_request_for_testing(tmp_path, request)
    assert response.status == "ok"

    monkeypatch.setattr(
        daemon_server.DaemonCore,
        "handle_request",
        lambda _self, _req: (_ for _ in ()).throw(
            ProtocolError("protocol version mismatch")
        ),
    )
    response = daemon_server.handle_request_for_testing(tmp_path, request)
    assert response.status == "error"
    assert response.error is not None
    assert response.error.code == "protocol_version_mismatch"

    monkeypatch.setattr(
        daemon_server.DaemonCore,
        "handle_request",
        lambda _self, _req: (_ for _ in ()).throw(RuntimeError("boom")),
    )
    response = daemon_server.handle_request_for_testing(tmp_path, request)
    assert response.status == "error"
    assert response.error is not None
    assert response.error.code == "internal_error"

    payload = json.dumps(request.model_dump(mode="json")).encode("utf-8")
    monkeypatch.setattr(
        daemon_server.DaemonCore,
        "handle_request",
        lambda _self, _req: ResponseEnvelope(
            protocol_version="1.0",
            request_id="req-1",
            status="ok",
            result={"status": "ok"},
        ),
    )
    raw_response = daemon_server.handle_raw_payload_for_testing(tmp_path, payload)
    assert raw_response.status == "ok"


def test_run_daemon_invokes_server_lifecycle(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    calls: list[str] = []

    class FakeServer:
        def __init__(self, _root: Path) -> None:
            calls.append("init")

        def __enter__(self):
            calls.append("enter")
            return self

        def __exit__(self, *_args):
            calls.append("exit")
            return False

        def warm_start(self) -> None:
            calls.append("warm")

        def serve_forever(self) -> None:
            calls.append("serve")

    monkeypatch.setattr(daemon_server, "DaemonServer", FakeServer)
    daemon_server.run_daemon(tmp_path)
    assert calls == ["init", "enter", "warm", "serve", "exit"]


def test_daemon_server_init_and_proxy_methods(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root = tmp_path
    socket_path = root / ".cache" / "kanbus.sock"
    socket_path.parent.mkdir(parents=True, exist_ok=True)
    socket_path.write_text("old", encoding="utf-8")
    monkeypatch.setattr(daemon_server, "get_daemon_socket_path", lambda _r: socket_path)
    captured: dict[str, object] = {}

    def fake_super_init(self, address, handler_cls):
        captured["address"] = address
        captured["handler"] = handler_cls

    monkeypatch.setattr(
        daemon_server.socketserver.ThreadingUnixStreamServer,
        "__init__",
        fake_super_init,
    )

    server = daemon_server.DaemonServer(root)
    assert isinstance(server.core, daemon_server.DaemonCore)
    assert captured["address"] == str(socket_path)
    assert captured["handler"] is daemon_server.DaemonRequestHandler
    assert not socket_path.exists()

    calls: dict[str, object] = {}
    monkeypatch.setattr(
        server.core, "warm_start", lambda: calls.setdefault("warm", True)
    )
    monkeypatch.setattr(
        server.core,
        "handle_request",
        lambda request: calls.setdefault("request", request)
        or ResponseEnvelope(
            protocol_version="1.0", request_id="req", status="ok", result={}
        ),
    )
    server.warm_start()
    req = RequestEnvelope(
        protocol_version="1.0", request_id="req", action="ping", payload={}
    )
    server.handle_request(req)
    assert calls.get("warm") is True
    assert calls.get("request") is req


def test_daemon_request_handler_handle_writes_response_and_shutdown(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    handler = object.__new__(daemon_server.DaemonRequestHandler)
    shutdown_called: dict[str, bool] = {}
    handler.server = type(
        "Server",
        (),
        {
            "core": object(),
            "shutdown": lambda self: shutdown_called.setdefault("shutdown", True),
        },
    )()
    handler.rfile = io.BytesIO(
        b'{"request_id":"r","action":"shutdown","protocol_version":"1.0","payload":{}}\n'
    )
    handler.wfile = io.BytesIO()

    response = ResponseEnvelope(
        protocol_version="1.0",
        request_id="r",
        status="ok",
        result={"status": "stopping"},
    )
    monkeypatch.setattr(
        daemon_server,
        "_handle_raw_request",
        lambda _core, _raw: (response, "shutdown"),
    )

    class FakeThread:
        def __init__(self, target, daemon):
            self.target = target
            self.daemon = daemon

        def start(self):
            self.target()

    monkeypatch.setattr(daemon_server.threading, "Thread", FakeThread)

    handler.handle()
    payload = handler.wfile.getvalue().decode("utf-8")
    assert '"status": "ok"' in payload
    assert shutdown_called.get("shutdown") is True

    # no-op when empty raw payload
    handler_empty = object.__new__(daemon_server.DaemonRequestHandler)
    handler_empty.server = handler.server
    handler_empty.rfile = io.BytesIO(b"")
    handler_empty.wfile = io.BytesIO()
    handler_empty.handle()
    assert handler_empty.wfile.getvalue() == b""


def test_handle_raw_request_protocol_error_branch() -> None:
    class ProtocolFailCore:
        def handle_request(self, _request):
            raise ProtocolError("protocol version unsupported")

    payload = json.dumps(
        {
            "protocol_version": "1.0",
            "request_id": "req-9",
            "action": "ping",
            "payload": {},
        }
    ).encode("utf-8")
    response, action = daemon_server._handle_raw_request(ProtocolFailCore(), payload)
    assert action == "ping"
    assert response.status == "error"
    assert response.error is not None
    assert response.error.code == "protocol_version_unsupported"



from __future__ import annotations

from pathlib import Path

import pytest

from kanbus import issue_close

from test_helpers import build_issue


def test_close_issue_delegates_to_update_with_closed_status(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    captured: dict[str, object] = {}
    expected = build_issue("kanbus-1", status="closed")

    def fake_update_issue(**kwargs):
        captured.update(kwargs)
        return expected

    monkeypatch.setattr(issue_close, "update_issue", fake_update_issue)

    result = issue_close.close_issue(tmp_path, "kanbus-1")

    assert result.identifier == "kanbus-1"
    assert result.status == "closed"
    assert captured["root"] == tmp_path
    assert captured["identifier"] == "kanbus-1"
    assert captured["status"] == "closed"
    assert captured["validate"] is True
    assert captured["claim"] is False


def test_close_issue_wraps_update_errors(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(
        issue_close,
        "update_issue",
        lambda **_kwargs: (_ for _ in ()).throw(
            issue_close.IssueUpdateError("update failed")
        ),
    )

    with pytest.raises(issue_close.IssueCloseError, match="update failed"):
        issue_close.close_issue(tmp_path, "kanbus-2")
